#!/usr/bin/env perl

#
#  Copyright (C) 2007 CeBiTec, Bielefeld University
#  Written by Wolfgang Gerlach.
#  The original version (CARMA 1.2) was developed by Lutz Krause.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#

use strict;
use DB;

use warnings;
no warnings 'recursion';

use FindBin;
use lib "$FindBin::Bin/../share/perl";
use File::Basename;
use File::Copy;

#use diagnostics;

use File::Glob ':glob';

use Getopt::Std;
use File::Temp qw(tempfile tempdir);
use FileUtils;
use TCommon;
#use TCarma_phylo;
#use TCarma_egt;

#use Bio::SearchIO;
use IO::Handle;

#use DNAUtils;
use TConfigs;





######################
# Configurations
######################


#######################################################################



$SIG{'INT'}  = 'handler';
$SIG{'TERM'} = 'handler';

my $INSTALLDIR = dirname($FindBin::Bin);
die "ERROR: INSTALLDIR variable not set in environment\n" unless $INSTALLDIR;

STDOUT->autoflush(1);





sub usage{
	print STDERR <<USAGE;
	
Copyright (C) 2007 CeBiTec, Bielefeld University.
Written by Lutz Krause.
This software comes with ABSOLUTELY NO WARRANTY; This is free
software, and you are welcome to redistribute it under
certain conditions.
Read the COPYRIGHT for details.


carma.pl: search and taxonomic classification of 
          short environmental DNA fragments
	
Usage: carma.pl <options> <fasta_input_file>
  where: fasta_file conains the reads of the studied environmental
	metagenome

	Options
	-p <c>  : project output directory (recommended)
	-o <c>*	: output file taxonomical classification 
                   (intermediate results are stored in temporary
                   directories, not recommended)
	-u <c>* : output file EGTs
		   * only needed when run without "-p"
	-c <n>	: submit <n> jobs
	-i	: submit on idle queue
		   (should ALWAYS be used on CeBiTec compute cluster)
	-e <n>	: hmmpfam evalue cut off
	-b <n>	: blast evalue cut off
	-h	: show this help message
	-t	: don't delete temporary files (for debugging)		 
	-v	: verbose


	old options, no idea whether they still work:
	(-f <n>	: only run for single family (e.g. PF05341))
	(-n <c>	: run for all families with name matching
			regular expression (PERL) <c>)
	(-g <n>	: run for all families annotated with certain GO-term
			(e.g. 'photosynthesis'))


USAGE
}

#########################
# ERROR handler, catches INT and TERM signal
#########################
sub handler {
	my $signal = shift;	# signal-nummer besorgen
	$SIG{'INT'}  = 'handler';   # reinstall sig-handler
	$SIG{'TERM'} = 'handler';   # reinstall sig-handler

	print STDERR "Signal: SIG$signal caught!\n";
	if ($signal eq "INT" or $signal eq "TERM"){
		print STDERR "Bailing out ...\n";
		die;
	}
	else{
		print STDERR "Caught signal $signal, don\'t know what to do !!!\n";
	}
	return 1;
}





##################
# Get options
##################
our($opt_v,$opt_e,$opt_h,$opt_b,$opt_o,$opt_c,$opt_m,$opt_j,$opt_u,
	$opt_s,$opt_p,$opt_D,$opt_i,$opt_f,$opt_g,$opt_n, $opt_t );

getopts('e:vhb:o:c:m:jsD:if:g:n:tu:p:');



#my $idle = $opt_i;




# print usage message
if($opt_h){
	usage;
	exit;
}

####################
# check parameters
####################
unless(@ARGV==1){
	usage;
	exit 1;
}

my $fasta_file=$ARGV[0];
unless(-e $fasta_file) {
	print STDERR "ERROR: Could not read file \"$fasta_file\"!\n";
	exit 1;
}


if ( (not defined $main::opt_o) && (not defined $main::opt_p) ) {
	print STDERR "ERROR: Please define either \"-p\" or \"-o\" !\n";
	exit 1;
}

if (defined $main::opt_o) {
	if (-e $main::opt_o) {

		print STDERR "ERROR: Output file -o \"".$main::opt_o."\" already exists!\n";
		exit 1;
	}
	
	my($filename, $directories, undef) = fileparse($main::opt_o);
	
	unless (-d $directories) {
		print STDERR "ERROR: Output directory \"$directories\" does not exist.\n";
		exit 1;
	}
	
}

if (defined $opt_u) {
	if (-e $main::opt_u) {

		print STDERR "ERROR: Output file -u \"".$main::opt_u."\" already exists!\n";
		exit 1;
	}
}


my $working_dir;

if ($opt_p) {
	unless ($opt_p =~ /^\//) {
		print STDERR "Please give absolute path name for option -p.\n";
		exit(1);
	}

	unless (-e $opt_p) {
		print STDERR "ERROR: Directory \"$opt_p\" does not exists.\n";
		exit(1);
	}
	
	unless (-d $opt_p) {
		print STDERR "ERROR: \"$opt_p\" is not a directoy.\n";
		exit(1);
	}
	
	$working_dir = $opt_p;

} else {
	#create a directory for the intermediate results
	$working_dir = tempdir ( DIR =>  TMP_DIR, CLEANUP => (($main::opt_t)?0:1));
}


my $working_dir_search   = $working_dir."/search/";	 # = tempdir ( DIR =>  $working_dir, CLEANUP => (($main::opt_t)?0:1));
my $working_dir_classify = $working_dir."/classify/";	# = tempdir ( DIR =>  $working_dir, CLEANUP => (($main::opt_t)?0:1));


unless (-e $working_dir_search) {
	unless (mkdir $working_dir_search) {
		print STDERR "ERROR: Error creating directory \"$working_dir_search\"!\n";
		exit(1);
	}
}

unless (-e $working_dir_classify) {
	unless (mkdir $working_dir_classify) {
		print STDERR "ERROR: Error creating directory \"$working_dir_classify\"!\n";
		exit(1);
	}
}



print "working_dir_search: $working_dir_search\n";
print "working_dir_classify: $working_dir_classify\n";


######################################################################################
# EGT Search

my $cmd=$FindBin::Bin."/carma_search.sh";
$cmd.= " -o ".$working_dir_search;

if ($opt_c) {
	$cmd.= " -c ".$opt_c;
}

if ($opt_e) {
	$cmd.= " -e ".$opt_e;
}

if ($opt_b) {
	$cmd.= " -b ".$opt_b;
}


if ($opt_i) {
	$cmd.= " -i";
}


if ($opt_v) {
	$cmd.= " -v";
}

if ($opt_t) {
	$cmd.= " -t";
}


$cmd.= " $fasta_file";



print STDERR "Call: \"".$cmd."\"\n";


unless (system($cmd) == 0) {
	print STDERR "ERROR: carma_search.pl stopped for some unknown reason.\n";
	exit(1);
}
print "carma_search.pl finished successfully!\n";

if ($opt_u) {
	$cmd="cat $working_dir_search/" . basename($fasta_file) . "/*.out > $opt_u";

	print "CMD: ".$cmd."\n";

	unless (system($cmd)) {
		print STDERR "Calling \"$cmd\" failed. Nevertheless I continue.\n"	;
	}
	
}


######################################################################################
# EGT Classification


sleep 2;

$cmd=$FindBin::Bin."/carma_classify.sh";
$cmd.= " -o ".$working_dir_classify;

if ($opt_c) {
	$cmd.= " -c ".$opt_c;
}

if ($opt_i) {
	$cmd.= " -i";
}


if ($opt_v) {
	$cmd.= " -v";
}

if ($opt_t) {
	$cmd.= " -t";
}


$cmd.= " $working_dir_search/".basename($fasta_file)."/";



print STDERR "Call: \"".$cmd."\"\n";

unless (system($cmd) == 0) {
	print STDERR "ERROR: carma_classify.pl stopped for some unknown reason.\n";
	exit(1);
}
print STDERR "carma_classify.pl finished successfully!\n";


######################################################################################
# Collect output and write to a single file

my $files_wildcard;
my @files;
my $result_file;

############## SEARCH Files:

if ((defined $opt_u) || (defined $opt_p) ) {

	$result_file = "$working_dir/result.egt";

	unless (-e $result_file) {
		$files_wildcard = "$working_dir_search/".basename($fasta_file)."/*.out";

		@files = glob($files_wildcard);

		unless (@files) {
			print STDERR "ERROR: No output files found !??";
			exit(1);
		}

		unless (open(OUTPUT, ">".$result_file)) {
			die $!;
		}

		foreach my $file (@files) {
			#print $file."\n";
			copy($file,\*OUTPUT);

		} 

		close OUTPUT;
	} else {
		print STDERR "File \"$result_file\" already exists, skipping... \n";
	}
	
}

if (defined $opt_u) {
	if (-e $opt_u) {
		print STDERR "File \"$opt_u\" already exists.";
		exit(1);
	} else {
		copy($result_file,$opt_u) or die "Copy failed: $!";
		print STDERR "Created File: \"$opt_u\"\n";
	}
}




############## CLASSIFICATION Files:


$result_file = "$working_dir/result.taxa";

unless (-e $result_file) {
	$files_wildcard = "$working_dir_classify/".basename($fasta_file)."/*.out";

	@files = glob($files_wildcard);

	unless (@files) {
		print STDERR "ERROR: No output files found !??";
		exit(1);
	}



	unless (open(OUTPUT, ">".$result_file)) {
		die $!;
	}

	foreach my $file (@files) {
		#print $file."\n";
		copy($file,\*OUTPUT);

	} 

	close OUTPUT;
} else {
	print STDERR "File \"$result_file\" already exists, skipping... \n";
}

if (defined $opt_o) {
	if (-e $opt_o) {
		print STDERR "File \"$opt_o\" already exists.";
		exit(1);
	} else {
		copy($result_file,$opt_o) or die "Copy failed: $!";
		print STDERR "Created File: \"$opt_o\"\n";
	}
}


######################################################################################
# Everything is done, now produce some nice pictures:

#### functional profile:
my $result_ps=$working_dir."/functional_profile.ps";
my $result_tsv=$working_dir."/functional_profile.tsv";

unless (-e $result_tsv) {
	$cmd=$FindBin::Bin."/../tools/getFunctionalProfile.pl";
	$cmd.= " -o $result_tsv -g $result_ps -s -l 40 $working_dir/result.egt"; #for gnuplot output

	unless (system($cmd) == 0) {
		print STDERR "ERROR: getFunctionalProfile.pl stopped for some unknown reason.\n";
		print STDERR "command was: \"$cmd\"\n";
		exit(1);
	}
} else {
	print STDERR "File \"$result_ps\" already exists, skipping... \n";
}

#### taxonomic profile:
$result_tsv=$working_dir."/taxonomic_profile.tsv";

#check only for kingdom.ps
unless (-e $result_tsv) {
	$cmd=$FindBin::Bin."/../tools/getTaxonomicProfile.pl";
	$cmd.= " -l 40 -o $result_tsv -g $working_dir $working_dir/result.taxa";

	unless (system($cmd) == 0) {
		print STDERR "ERROR: getTaxonomicProfile.pl stopped for some unknown reason.\n";
		print STDERR "command was: \"$cmd\"\n";
		exit(1);
	}
} else {
	print STDERR "(At least) \"$result_tsv\" already exists, skipping... \n";
}

##### convert ps2pdf


unless (-e "$working_dir/superkingdom.pdf") {

	my $ps2pdf = `which ps2pdf`;
	chomp($ps2pdf);

	if ($ps2pdf) {

		$files_wildcard = "$working_dir/*.ps";

		@files = glob($files_wildcard);
	
		foreach my $file (@files) {
			my $outputfile = $file;
			$outputfile =~ s/ps$/pdf/;
			$cmd = "$ps2pdf $file $outputfile";
			if (system($cmd) != 0) {
				print STDERR "ERROR calling \"$cmd\", returned: $?\n";
				exit(1);
			}

		}
		
		#delete postscripts, keep only pdf:
		system ("rm $working_dir/*.ps");
		
	} else {
		print STDERR "ERROR: ps2pdf not found\n";
	}
	
	

} else {
	print STDERR "(At least) \"$working_dir/superkingdom.pdf\" already exists, skipping... \n";
}




print STDERR "carma.pl finished successfully!\n";


exit(0);

