#!/usr/bin/env perl


use strict;

use warnings;
use TCommon;
use File::Basename;
use Getopt::Std;
use POSIX qw(tmpnam);
use File::Copy;


sub usage{
    print STDERR <<USAGE;
    
Get seperate GO-term profile for diffierent taxonomic groups

Usage: getGOprofile.pl [-p] <directory> <T1> <T2> ... <Tn>

  where: <directory> is a directory storing results of taxonomic classification 
                     of reads
	 <Ti>        compute GO-term profile for taxonomic group
                     (e.g. <Ti> = 'Bacteria_Firmicutes_Clostridia')
	 -r          print ratios (default is counts)
	 -m <i>      mincount

USAGE
    
}

sub pfam_family_counts{
    my ($query,$hits,$count,$evalue_cf,$total) = @_;

    foreach my $hit (@$hits){
	my ($acc,$descr,$score,$evalue) = @$hit;

	$$total++;
	
	unless(exists $count->{$acc}){
	    $count->{$acc} = [$descr,1];
	}
	else{
	    $count->{$acc}->[1]++;
	}
    }
    return 1;
}



our($opt_v,$opt_h,$opt_m, $opt_t);

getopts('vhpm:t');

unless(@ARGV >= 2){
    usage;
    exit 1;
}

if($opt_h){
    usage;
    exit;
}


my $indir = shift @ARGV;

# Hash storing all taxonomic groups
# for which GO-term profile is computed
my %taxa = ();
foreach (@ARGV){
    $taxa{$_} = 1;
}



print STDERR "Reading directory $indir ...";
opendir DIR, "$indir" || die "ERROR: Could not open directory $indir:$!\n";

my @allentries = grep /\.out$/, grep !/^\.\.?$/, readdir DIR;
close DIR;
print STDERR "Done.\n";

my $number = @allentries;

my %go_counts = ();

my %total = ();

my $n = 0;
print STDERR "Parsing $number files ...\n";
foreach my $file (@allentries){
    $n++;
    $file = "$indir/$file";
    
    print STDERR "\r$n/$number";
    
    die "ERROR: could not open file $file:\n$!\n" unless open( FH, "<$file" );
    my $line;
    
    while ( $line = <FH> ) {
	next if $line =~ /^\s*$/;
	my @tmp = split( /\s+/, $line );
	die "ERROR: $line\n" unless @tmp == 3;
	
	my ($pfam_acc,$id,$t) = @tmp;
	
	$t =~ s/\;//g;
	    
	$t = 'Unknown' if $t eq '';
	    
	if($t =~ /(_[a-zA-Z]+:[a-zA-Z]+)/){
	    my $species = $1;
	    die "$species\n" unless $species =~ /^(_[a-zA-Z]+)/;
	    my $genus = $1;
	    my $species_new = $species;
	    
	    unless ($species_new =~ s/$genus\:/${genus}$genus\:/){
	      die "$species_new  $genus\n";
	    }
	    my $new_taxa = $t;
	    die unless $new_taxa =~ s/$species/$species_new/;
	    $t = $new_taxa unless $new_taxa =~ /$genus$genus$species/;
         }
         
         my $match = 0;
         foreach my $taxa (keys %taxa){

	     next unless $t =~ /^$taxa/;
	     $match = $taxa;
	     last;
         }

        if($match){
	    my $pfam_id = TCommon::get_pfamID_by_pfamACC($pfam_acc);
	    my $go = TCommon::get_go_by_pfam_name($pfam_id);
	    
	    next unless @$go;

	    foreach my $g (@$go){
		my ($term,$class) = ($g->{term},$g->{class});

		$go_counts{$class}->{$term}->{$match}++;
		$total{$match}++;
	    }
	    
	}
    }
    close(FH);
    
}


print STDERR " Parsing completed.\n";

my @taxa_keys = keys %taxa;

print ',,';
foreach (@taxa_keys){
    print "$_,";
}
print "\n";

print "Class,Term,";
foreach (@taxa_keys){
    print "Number of hits,";
}
print "\n";

foreach my $class (sort keys %go_counts){
    foreach my $term (sort {$go_counts{$class}->{$b} <=> $go_counts{$class}->{$a}} 
		      keys %{ $go_counts{$class} }){


	if($opt_m){
	    my $max = 0;
	    foreach my $t (@taxa_keys){
		my $count = $go_counts{$class}->{$term}->{$t} || 0;
		$max = $count if $count > $max;
	    }
	    next unless $max > $opt_m;
	}

	print "$class,$term,";
	foreach my $t (@taxa_keys){
	    
	    my $count = $go_counts{$class}->{$term}->{$t} || 0;

	    if($opt_t){
		$count = $count / $total{$t};
	    }
	    print "$count,";
	}
	print "\n";
    }
}


print "\n\nTotal counts:";

foreach my $taxa (@taxa_keys){
    print "$taxa: " , $total{$taxa} , "\n";
}


