#!/usr/bin/env perl

#
#  Copyright (C) 2007 CeBiTec, Bielefeld University
#  Written by Lutz Krause (Lutz.Krause@CeBiTec.Uni-Bielefeld.DE)
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#

use strict;

use warnings;
use Getopt::Std;
use File::Basename;
use POSIX qw(tmpnam);
use TCommon;

use Bio::SearchIO;

use IO::Handle;

use DNAUtils;

use TConfigs;


######################
# Configurations
######################

use constant MAX_FAMILY_SIZE => 3000;
use constant BLAST_DEFAULT_EVALUE => 1;
use constant HMMPFAM_DEFAULT_EVALUE => 0.01;

#######################################################################

my $OUTFILE = '';

$SIG{'INT'}  = 'handler';
$SIG{'TERM'} = 'handler';

my $INSTALLDIR = $ENV{INSTALLDIR};

die "ERROR: INSTALLDIR variable not set in environment\n" unless $INSTALLDIR;

STDOUT->autoflush(1);

sub usage{
    print STDERR <<USAGE;
    
Copyright (C) 2007 CeBiTec, Bielefeld University.
Written by Lutz Krause.
This software comes with ABSOLUTELY NO WARRANTY; This is free
software, and you are welcome to redistribute it under
certain conditions.
Read the COPYRIGHT for details.


process_alignments.pl: Process Pfam family multiple alignments:

Usage: process_alignments.pl <Pfam-A.full> 
  where: <Pfam-A.full> the full alignments of the 
                       curated Pfam families

    Options
	-h     : show this help message
        -v     : verbose

USAGE
}

##############################
# Processes multiple alignment
# Removes duplicate entries and
# members without known taxonomic
# affiliation
# Results are printed in fasta
# format into directory
# out_dir
##############################
sub process_alignment{
    my ($family_name,$alignment,$out_dir,$warnings,$duplicates) = @_;

    print "$family_name\n";

    my $out_file = "$out_dir/$family_name.fas";
    
    my %seq_by_acc = ();
    my %proc = ();
    my @alignment_processed = ();

    # Remove duplicate entries
    #
    foreach (@$alignment){
	my ($acc,$seq) = @$_;
	
	my $taxa = TCommon::get_taxa($acc);
	
	# skip members without known taxonomic affiliation
	unless (ref $taxa and @$taxa) {
	    print STDERR "ERORR: Could not get taxa for acc $acc\n" if WARNINGS;
	    $$warnings++;
	    next;
	}

	$taxa = join( '_', @$taxa );

	# skip environmental sequences
	if ( $taxa =~ /environmental/i ) {
	    next;
	}
	
	# some checks
	if ( exists $seq_by_acc{$acc} ) {	    
	    print STDERR "\n$acc\n$seq\n\n";
	    print STDERR $seq_by_acc{$acc}, "\n";
	    die;
	}

	$seq_by_acc{$acc} = $seq;

	# remove duplicates
	if ( exists $proc{$taxa} ) {
	    my $rem = 0;
	    foreach my $s ( @{ $proc{$taxa} } ) {
		if ( $s eq $seq ) {
		    $$duplicates++;
		    $rem = 1;
		    last;
		}
	    }
	    next if $rem;
	    push @{ $proc{$taxa} }, $seq;
	}
	else {
	    $proc{$taxa} = [$seq];
	}
	
	die if $seq =~ /\?/;
	
	push @alignment_processed, $_;
    }
    
    return 1 unless @alignment_processed;

    # Print results in fasta format
    #
    unless(open(OUT,">$out_file")){
	print STDERR "ERROR: Could not open file $out_file:\n$!\n";
	exit 1;
    }

    foreach (@alignment_processed){
	my ($n,$s) = @$_;

	$s = uc($s);
	$s =~ s/\./\-/g;
	

	print OUT ">$n\n$s\n";
    }

    close(OUT);

    return 1;
}


#########################
# ERROR handler, catches INT and TERM signal
#########################
sub handler {
    my $signal = shift;    # signal-nummer besorgen
    $SIG{'INT'}  = 'handler';   # reinstall sig-handler
    $SIG{'TERM'} = 'handler';   # reinstall sig-handler

    print STDERR "Signal: SIG$signal caught!\n";
    if ($signal eq "INT" or $signal eq "TERM"){
	print STDERR "Bailing out ...\n";
	die;
    }
    else{
	print STDERR "Caught signal $signal, don\'t know what to do !!!\n";
    }
    return 1;
}

sub dir_is_empty
{
    my ($path) = @_;
    opendir DIR, $path;
    while(my $entry = readdir DIR) {
        next if($entry =~ /^\.\.?$/);
        closedir DIR;
        return 0;
    }
    closedir DIR;
    return 1;
}


##################
# Get options
##################
our($opt_v,$opt_h);

getopts('vh');

# print usage message
if($opt_h){
    usage;
    exit;
}

####################
# check parameters
####################
unless(@ARGV == 1){
    usage;
    
    exit 1;
}

# Process full alignments of curated
# Pfam families:
# 1) remove duplicate entries
# 2) transform stockholm into fasta format
# 3) remove members without known taxonomic affiliation

my ($pfam_a_full) = @ARGV;

my $out_dir = PFAM_FAMILY_ALIGNMENTS;

unless(dir_is_empty($out_dir)) {
	die "$out_dir should be empty to avoid overwriting.\n";
}


my ($warnings,$duplicates) = (0,0);

unless(TCommon::parse_stockholm($pfam_a_full,\&process_alignment,
				$out_dir,\$warnings,\$duplicates)){
    print STDERR "ERROR while parsing file $pfam_a_full\n";
    exit 1;
}

print STDERR "$warnings warnings\n";
print STDERR "Removed $duplicates duplicates\n";
