package DNAUtils;
#
#  This file is part of GenDB 2.2
#  Copyright (C) 2005 CeBiTec, Bielefeld University
#                     <gendb@CeBiTec.Uni-Bielefeld.DE>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#
=head1 NAME

DNAUtils

=head1 DESCRIPTION

various functions and methods to handle DNA sequences

=head1 SYNOPSIS

use DNAUtils;
use Carp;

my $dna_seq = "ATGTGCATGATGCTAAGAT...";

# object oriented interface,
# using genetic code 4
my $dnautils = DNAUtils->new(4);
my $aa_seq = $dnautils->translate($dna_seq);

# static interface
# set global genetic code to 4
DNAUtils->set_genetic_code(4);
$aa_seq = DNAUtils->translate($dna_seq);

# static interface, using imported functions
$aa_seq = translate($dna_seq);

# passing genetic code as additional parameter
$aa_seq = translate($dna_seq, 11);

=cut

    use strict;
use warnings;
use DNAUtils::Codes qw(CODON_TABLES START_CODONS STOP_CODONS);
use base qw(Exporter);

our @EXPORT = qw(set_genetic_code
		 get_genetic_code
		 reverse_complement
		 complement
		 dna2aa
		 translate
		 translate_6_frames
		 is_startcodon
		 is_stopcodon);

# the currently global genetic code
# (used for all functions without an
# object or an explicitly given code)
# defaults to standard bacteria code
our $global_genetic_code = 11;

1;

=head1 Methods and Function

DNAUtils provides an object oriented interface for all methods and functions. The genetic code used inside an object is given to the constructor. Most of the methods and functions may also be called statically.

THE STATIC INTERFACE IS DEPRECATED AND WILL BE REMOVED IN THE FUTURE !

=head2 Management of genetic code

The DNAUtils module supports a subset of the genetic codes. The following methods and functions allow an application to set and retrieve it.


=over 4

=item * DNAUtils->new([genetic code])

creates a new DNAUtils object. If no genetic code is given, the default code (1) is used. If an unsupported code is given, the constructor returns undef instead of an object.

=cut

sub new {
    my ($class, $code) = @_;
    $code = 11 unless ($code);

    return undef unless (defined (CODON_TABLES->{$code}));
    my $self = {code => $code};
    return bless $self, ref $class || $class;
}

=item * $dnautils->genetic_code()

returns the genetic code of the given DNAUtils object.

=cut

sub genetic_code {
    return $_[0]->{code};
}


=item * DNAUtils->get_genetic_code_list()

returns an array containing the available genetic code.

=cut

sub get_genetic_code_list {
    return keys(%{CODON_TABLES()});
}

=item * DNAUtils->set_genetic_code(code)   [DEPRECATED]

sets the global genetic code (used as default code for statically invoked methods)

=cut

sub set_genetic_code {
    my $code;
    if (scalar(@_) == 2) {
	$code = $_[1];
    }
    else {
	$code = $_[0];
    }
    
    if (defined($code)) {
	if (defined (CODON_TABLES()->{$code})) {
	    $global_genetic_code = $code;
	}
	else {
	    print STDERR "unknown genetic code $code ".
		"(actual set code: $global_genetic_code)\n";
	}
    } 
    else {
	print STDERR "set_genetic_code: undefined parameter\n";
    }
}


=item * DNAUtils->get_genetic_code()   [DEPRECATED]

returns the current global genetic code

=cut

sub get_genetic_code {
    return $global_genetic_code;
}

=back

=head2 DNA translation

=over 4

=item * translate(dna sequence [,genetic code])

splits the given dna sequence in codon and returns the translation according tothe genetic code to be used. The genetic code may be passed as additional parameter or is taken from the object if called as a method.

Unknown codons or codons with invalid character (no A, T, G or C) are translated to "X".

=cut

sub translate {
    my ($seq, $genetic_code);
    if (ref($_[0])) {
	$seq = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $_[0]->genetic_code();
    }
    elsif ($_[0] eq __PACKAGE__) {
	$seq = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $global_genetic_code;
    }
    else {
	$seq = $_[0];
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }

    if (defined (CODON_TABLES()->{$genetic_code})) {
	my $aa = "";
	my $num = length($seq)/3;
	foreach (unpack( "a3" x $num, uc($seq))) {
	    if (defined (CODON_TABLES()->{$genetic_code}) &&
		defined (CODON_TABLES()->{$genetic_code}->{$_})) {
		$aa .= CODON_TABLES()->{$genetic_code}->{$_};
	    }
	    elsif($seq =~/^\s*$/){
		print STDERR "ERROR: sequence contains new lines!";
	    }
	    else{
		$aa .= 'X';
	    }
	}
	return $aa;
    }
    else {
	return 'X' x (int(length($seq)/3));
    }
}

=item * dna2aa(codon [,genetic_code])

translates a single codon into its amino acid

=cut

sub dna2aa {
    my ($codon, $genetic_code);
    if (ref($_[0])) {
	$codon = uc($_[1]);
	$genetic_code = ($_[2]) ? $_[2] : $_[0]->genetic_code();
    }
    elsif ($_[0] eq __PACKAGE__) {
	$codon = uc($_[1]);
	$genetic_code = ($_[2]) ? $_[2] : $global_genetic_code;
    }
    else {
	$codon = uc($_[0]);
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }
    if (defined (CODON_TABLES()->{$genetic_code}) &&
	defined (CODON_TABLES()->{$genetic_code}->{$codon})) {
	return CODON_TABLES()->{$genetic_code}->{$codon};
    }
    else {
	return 'X';
    }
}

=item * translate_6_frames(dna sequence [,genetic code])

returns the translation of all 6 reading frames

=cut 

sub translate_6_frames {
    my ($seq, $genetic_code);
    if (ref($_[0])) {
	$seq = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $_[0]->genetic_code();
    }
    elsif ($_[0] eq __PACKAGE__) {
	$seq = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $global_genetic_code;
    }
    else {
	$seq = $_[0];
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }

    if (defined (CODON_TABLES()->{$genetic_code})) {
	my $length = length($seq);
	my @results;
	foreach my $offset (0..2) {
	    push @results, translate(substr($seq, $offset), $genetic_code);
	}
	my $rev_seq = reverse_complement($seq);
	foreach my $offset (0..2) {
	    push @results, translate(substr($rev_seq, $offset), $genetic_code);
	}
	return \@results;
    }
    else {
	return [];
    }
}

=back

=head2 Start/Stop codons

=over 4

=item * get_start_codons([genetic code])

returns a list of all start codons in the active genetic code (or the code given as parameter.

=cut

sub get_start_codons {
    my $genetic_code;
    if (defined($_[0]) && ref($_[0])) {
	$genetic_code = ($_[1]) ? $_[1] : $_[0]->genetic_code();
    }
    elsif (defined($_[0]) && ($_[0] eq __PACKAGE__)) {
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }
    else {
	$genetic_code = ($_[0]) ? $_[0] : $global_genetic_code;
    }
    if (defined(START_CODONS->{$genetic_code})) {
	return @{START_CODONS->{$genetic_code}} if (wantarray);
	return START_CODONS->{$genetic_code};
    }
    return () if (wantarray);
    return [];
}

=item * get_stop_codons([genetic code])

returns a list of all stop codons in the active genetic code (or the code given as parameter.

=cut

sub get_stop_codons {
    my $genetic_code;
    if (defined($_[0]) && ref($_[0])) {
	$genetic_code = ($_[1]) ? $_[1] : $_[0]->genetic_code();
    }
    elsif (defined($_[0]) && $_[0] eq __PACKAGE__) {
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }
    else {
	$genetic_code = ($_[0]) ? $_[0] : $global_genetic_code;
    }
    if (defined(STOP_CODONS->{$genetic_code})) {
	return @{STOP_CODONS->{$genetic_code}} if (wantarray);
	return STOP_CODONS->{$genetic_code};
    }
    return () if (wantarray);
    return [];
}

=item * is_startcodon(codon [,genetic code])

returns true if the given codon is a start codon

=cut 

sub is_startcodon {
    my ($codon, $genetic_code);
    if (ref($_[0])) {
	$codon = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $_[0]->genetic_code();
    }
    elsif ($_[0] eq __PACKAGE__) {
	$codon = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $global_genetic_code;
    }
    else {
	$codon = $_[0];
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }
    $codon =~ tr/atcg/ATCG/;

    foreach (get_start_codons($genetic_code)) {
	return 1 if ($_ eq $codon);
    }
    return 0;
}

=item * is_stopcodon(codon [,genetic code])

returns true if the given codon is a start codon

=cut 

sub is_stopcodon {
    my ($codon, $genetic_code);
    if (ref($_[0])) {
	$codon = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $_[0]->genetic_code();
    }
    elsif ($_[0] eq __PACKAGE__) {
	$codon = $_[1];
	$genetic_code = ($_[2]) ? $_[2] : $global_genetic_code;
    }
    else {
	$codon = $_[0];
	$genetic_code = ($_[1]) ? $_[1] : $global_genetic_code;
    }
    
    $codon =~ tr/atcg/ATCG/;
    foreach (get_stop_codons($genetic_code)) {
	return 1 if ($_ eq $codon);
    }
    return 0;
}

=back

=head2 Sequence handling

These functions work indepently of the genetic code.

=over 4

=item * complement(dna sequence)

returns the complement of the given sequence

=cut

sub complement {
    my $dna_seq = lc((scalar(@_) == 2) ? $_[1] : $_[0]);
    $dna_seq =~ tr/agctn/tcgan/;
    return $dna_seq;
}

=item * reverse_complement(dna sequence)

returns the reverse complement of the given sequence

=cut

sub reverse_complement {
    my $dna_seq = reverse(lc((scalar(@_) == 2) ? $_[1] : $_[0]));
    $dna_seq =~ tr/agctn/tcgan/;
    return $dna_seq;
}


=item * is_startcodon(codon [,genetic code])

returns ref on hash on hash on array of ALL (not only longest) ORFs of a sequence

=cut 


sub get_orfs_all{
    my ($dnaseq,$orf_min_len,$verbose) = @_;

    die unless defined $orf_min_len;

    my $orfs = {};
    my $orf_number;


    for my $frame (1,2,3){
	my ($frame_orfs,$frame_orf_number) = _get_orfs_frame_forward_all($dnaseq,$frame,$orf_min_len);


	$orfs->{$frame} = $frame_orfs;
	$orf_number += $frame_orf_number;
    }


    for my $frame (-1,-2,-3){
	my ($frame_orfs,$frame_orf_number) = _get_orfs_frame_reverse_all($dnaseq,$frame,$orf_min_len);


	$orfs->{$frame} = $frame_orfs;
	$orf_number += $frame_orf_number;
    }    


    return ($orfs,$orf_number);  
}


# returns ref on hash on hash on array of ALL (not only longest) ORFs of a sequence on forward strand
sub _get_orfs_frame_forward_all {
    my ($dnaseq,$frame,$orf_min_len) = @_;


    # get the DNA sequence of the frame
    $dnaseq = substr($dnaseq,$frame - 1) if $frame > 1;


    my $codon_number = int(length($dnaseq) / 3);


    # get all codons of the frame
    my @codon_array = unpack("a3" x $codon_number, $dnaseq);


    my $orf_number = 0;


    my $orfs = {};


    # position: position in the DNA sequence, hence take $frame into account
    my $position = -3 + $frame;


    my @start_codons = ();


    foreach my $codon (@codon_array) {
	$position += 3;


	if(&is_stopcodon($codon) == 1){
	    my $stop = $position + 2; 


	    if (@start_codons > 0){


		foreach my $st (@start_codons){
		    if(($stop - $st) >= $orf_min_len){
			push @{ $orfs->{$stop} }, $st;
			$orf_number++;
		    }
		}
	    }


	    @start_codons = ();
	    next;
	}
	if(&is_startcodon($codon) == 1){
	    push @start_codons, $position;
	}
    }
    return ($orfs,$orf_number);   
}

# returns ref on hash on hash on array of ALL (not only longest) ORFs of a sequence on reverse strand
sub _get_orfs_frame_reverse_all{
    my ($dnaseq,$frame,$orf_min_len) = @_;


    my $absframe = abs($frame);


    # get the DNA sequence of the frame
    $dnaseq = substr($dnaseq,$absframe - 1) if $absframe > 1;


    my $codon_number = int(length($dnaseq) / 3);


    # get all codons of the frame
    my @codon_array = unpack("a3" x $codon_number, $dnaseq);


    my $last_stop;
    my $orf_number = 0;


    my $orfs = {};


    # position: position in the DNA sequence, hence take $frame into account
    my $position = -3 + $absframe;


    foreach my $codon (@codon_array) {
	$position += 3;


	$codon = reverse_complement($codon);


	if(&is_stopcodon($codon) == 1){
	    $last_stop = $position;
	    next;
	}
	if(&is_startcodon($codon) == 1){
	    my $start = $position + 2;
	    if (defined $last_stop and ($start - $last_stop) >= $orf_min_len ){		
		push @{ $orfs->{$last_stop} }, $start;
		die "stop: $position start: $start \n" if $start < 0;
		($start - $position + 1) % 3 and die "$frame, $start, $position";
		$orf_number++;
	    }
	}
    }
    return ($orfs,$orf_number);       
}


=item * get_orfs{(seq, orf_min_len [,verbose])

extracts all ORFs of a given sequence. Returns tupel
(number of found ORFs, ref on Hash of Hash containing all ORFs)

=cut 

sub get_orfs{
    my ($dnaseq,$orf_min_len,$verbose) = @_;

    die unless defined $orf_min_len;

    my $orfs = {};
    my $orf_number;


    for my $frame (1,2,3){
	my ($frame_orfs,$frame_orf_number) = _get_orfs_frame_forward($dnaseq,$frame,$orf_min_len);


	$orfs->{$frame} = $frame_orfs;
	$orf_number += $frame_orf_number;
    }


    for my $frame (-1,-2,-3){
	my ($frame_orfs,$frame_orf_number) = _get_orfs_frame_reverse($dnaseq,$frame,$orf_min_len);


	$orfs->{$frame} = $frame_orfs;
	$orf_number += $frame_orf_number;
    }


    return ($orfs,$orf_number);    
}

# returns ref on hash on hash on array of all longest ORFs of a sequence on forward strand
sub _get_orfs_frame_forward{
    my ($dnaseq,$frame,$orf_min_len) = @_;


    # get the DNA sequence of the frame
    $dnaseq = substr($dnaseq,$frame - 1) if $frame > 1;


    my $codon_number = int(length($dnaseq) / 3);


    # get all codons of the frame
    my @codon_array = unpack("a3" x $codon_number, $dnaseq);


    my ($last_start);


    my $orf_number = 0;


    my $orfs = {};


    # position: position in the DNA sequence, hence take $frame into account
    my $position = -3 + $frame;


    foreach my $codon (@codon_array) {
	$position += 3;


	if(&is_stopcodon($codon) == 1){
	    my $stop = $position + 2; 


	    if (defined $last_start and ($stop - $last_start) >= $orf_min_len ){		
		$orfs->{$stop} = $last_start;
		$orf_number++;
	    }


	    $last_start = undef;
	    next;
	}
	if((&is_startcodon($codon) == 1) and ! defined $last_start){
	    $last_start = $position;
	}
    }
    return ($orfs,$orf_number);   


}

# returns ref on hash on hash on array of all longest ORFs of a sequence on reverse strand
sub _get_orfs_frame_reverse{
    my ($dnaseq,$frame,$orf_min_len) = @_;

    my $absframe = abs($frame);


    # get the DNA sequence of the frame
    $dnaseq = substr($dnaseq,$absframe - 1) if $absframe > 1;


    my $codon_number = int(length($dnaseq) / 3);


    # get all codons of the frame
    my @codon_array = unpack("a3" x $codon_number, $dnaseq);


    my ($last_stop,$last_start) = (undef, undef);


    my $orf_number = 0;


    my $orfs = {};


    # position: position in the DNA sequence, hence take $frame into account
    my $position = -3 + $absframe;


    foreach my $codon (@codon_array) {
	$position += 3;


	$codon = reverse_complement($codon);


	if(&is_stopcodon($codon) == 1){
	    if (defined $last_start and defined $last_stop and ($last_start - $last_stop) >= $orf_min_len ){		


		$orfs->{$last_stop} = $last_start;
		die "stop: $position start: $last_start \n" if $last_start < 0;
		($last_start - $position + 1) % 3 and die "$frame, $last_start, $position";
		$orf_number++;
		}
	    $last_stop = $position;
	    $last_start = undef;
	    next;
	}
	if(&is_startcodon($codon) == 1){
	    $last_start = $position + 2;
	}
    }


    # get orf at the 3' end
    if (defined $last_start and defined $last_stop and ($last_start - $last_stop) >= $orf_min_len ){			
	$orfs->{$last_stop} = $last_start;
	die "stop: $position start: $last_start \n" if $last_start < 0;
	($last_start - $position + 1) % 3 and die "$frame, $last_start, $position";
	$orf_number++;
    }


    return ($orfs,$orf_number);   


}

=item * translate_for_hmmer_pfam_search{(seq [,genetic code])

translates DNA sequence to AA sequence for hmmpfam search. * will be
replaced by Z

=cut 

sub translate_for_hmmer_pfam_search{
    my ($seq, $genetic_code)=@_;
    $genetic_code = get_genetic_code unless ($genetic_code);


    my $aa = translate($seq, $genetic_code);    
    $aa =~ tr/*/Z/;
    return $aa;
};

=item * is_orf(sequence)

checks if given sequence is valid ORF

=cut 

sub is_orf{
    my $seq = shift;

    my $has_stop = 0;

    my $codons = int(length($seq)/3);
    my @codon_array = unpack("a3" x $codons, $seq);

    my $i = 0;
    foreach my $codon (@codon_array) {
	$i++;

	if ($i == 1){
	    unless(&DNAUtils::is_startcodon($codon) == 1){
		print STDERR  "first codon not start codon ...";
		return;
	    }
	}
	if ($i == $codons){
	    unless(&DNAUtils::is_stopcodon($codon) == 1){
		print STDERR  "last codon not stop codon ... $codon $i ";
		return;
	    }
	    $has_stop = 1;	    
	    last;
	}
	if(&DNAUtils::is_stopcodon($codon) == 1){
	    print STDERR  "In frame stop codon !!!\n";
	    return;
	}
    }
    $has_stop or die;
    return 1;
}


=back
