package Scheduler::DRMAA::Replicator;
#
#  This file is part of GenDB 2.2
#  Copyright (C) 2005 CeBiTec, Bielefeld University
#                     <gendb@CeBiTec.Uni-Bielefeld.DE>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#
=head1 NAME

Scheduler::DRMAA::Replicator

=head1 DESCRIPTION

scheduling of GenDB jobs using a DRMAA-compliant interface and support for Replicator file transfers

=cut

use strict;
use warnings;
use Scheduler qw(:JOB_FLAGS);
use base qw(Scheduler);

use IO::File;
use File::Glob qw(:glob);
use File::Temp;
use File::Which qw();

# list of directories containing files under the control
# of replicator (including their replicator target dirs)
use constant REPLICATOR_DIRS => { '/vol/biodb/asn1' => '/usr/local/replic/SYNC',
				  '/vol/biodb/fasta' => '/usr/local/replic/SYNC'};

# the pathname of the replic-submit binary
use constant REPLICATOR_SUBMIT => '/usr/local/replic/BIN/replic-submit';

# the gateway host the meta language processor runs on
use constant REPLICATOR_GATEWAY => 'doc127';

# options for submitting replic-submit calls
# (e.g. special queues/complexes or environments)
use constant QSUB_OPTIONS => '-cwd -l rep_sub=1 -v LD_LIBRARY_PATH=/vol/codine-6.2/lib/sol-sparc';

our ($qsub_binary);

$qsub_binary = undef;

1;

# this scheduler does not really support asynchronous operations,
# but the necessary functions are available nonetheless
sub support_asynchronous {
    return 1;
}

sub job_based {
    return 0;
}

sub new {
    my ($class, %parameters) = @_;
    my $self = $class->SUPER::new(%parameters);
    bless $self, ref $class || $class;
    
    $self->{job_info} = {};
    return $self;
}

sub available {
    unless (defined ($qsub_binary)) {
	# try to locate the qsub binary from Grid Engine
	$qsub_binary = File::Which::which('qsub') || '';
    }
    # this scheduler is only usable if qsub is found in the path
    return ($qsub_binary ne '');
}

sub set_native_option {
    my ($self, $value) = @_;
    $self->{_native_options} = $value;
}

sub submit {
    my ($self) = @_;
    
    my @names = (keys %{$self->{sequences}}, keys %{$self->{data_sets}});

    # prepare the input files
    my $counter = 0;
    foreach (@names) {
	$counter++;
	# create the input and output files
	my $input_file = $self->{temp_dir}."/$counter.dat";
	if (exists($self->{sequences}->{$_})) {
	    # write sequences as fasta file
	    open (INPUT, '>', $input_file) || die "unable to open file $input_file for writing: $!";
	    print INPUT ">$_\n";
	    print INPUT $self->{sequences}->{$_};
	    print INPUT "\n";
	    close (INPUT);
	}
	else {
	    # plain data file
	    open (INPUT, '>', $input_file) || die "unable to open file $input_file for writing: $!";
	    binmode INPUT;
	    print INPUT $self->{data_sets}->{$_};
	    close (INPUT);
	}
	$self->{job_info}->{$_}->{input} = $input_file;
	$self->{job_info}->{$_}->{output} = $self->{temp_dir}."/$counter.stdout";
    }
    
    # since replicator has a limit on the length of qsub command line
    # options, we use an external file to store the options
    my $sge_options_file = File::Temp->new(TEMPLATE => 'options_XXXXXXXX',
					   SUFFIX => '.sge',
					   DIR => $self->{temp_dir});
    if ($self->{error_file}) {
	print $sge_options_file '-e '.$self->{error_file}."\n";
    }
    else {
	print $sge_options_file "-e /dev/null\n";
    }

    # redirect stdin and stdout to the data files
    print $sge_options_file '-i '.$self->{temp_dir}.'/$TASK_ID.dat'."\n";
    print $sge_options_file '-o '.$self->{temp_dir}.'/$TASK_ID.stdout'."\n";

    # add further options given by the user
    print $sge_options_file $self->{_native_options}."\n" if ($self->{_native_options});

    # run in current working directory
    print $sge_options_file "-cwd\n";
    close $sge_options_file;
    
    # set the command line
    my ($command, @arguments) = @{$self->{commandline}};
    
    @arguments = () unless (@arguments);
    if (scalar(@arguments)) {
	# substitute the input/output placeholders in the arguments
	if (defined ($self->{input_placeholder})) {
	    @arguments = map { ($_ && $_ eq $self->{input_placeholder}) ? 
				   '/dev/fd/0' : $_} @arguments;
	}
	if (defined ($self->{output_placeholder})) {
	    # set the output place holder to /dev/fd/1 aka stdout
	    @arguments = map { ($_ && $_ eq $self->{output_placeholder}) ? 
				   '/dev/fd/1' : $_} @arguments;
	}
    }

    # instead of submitting the command itself, we create a replicator control
    # file which handles the transfer of files prior to starting the jobs
    
    my @transfer_files;
    # scan the arguments for directories in the replicator dir list
    for (my $i = 0; $i < scalar(@arguments); $i++) {
	if (my ($dir) = grep {index($arguments[$i], $_) == 0} keys %{REPLICATOR_DIRS()}) {
	    # find the files matching to the given argument and put them into
	    # the transfer list
	    push @transfer_files, bsd_glob($arguments[$i].'*');

	    # replace the directory part which the replicator dir
	    # if necessary
	    if (REPLICATOR_DIRS->{$dir} ne $dir) {
		substr($arguments[$i],0,length($dir)) = REPLICATOR_DIRS->{$dir};
	    }
	}
    }
    
    my $td_file = File::Temp->new(TEMPLATE => 'control_XXXXXXXX',
				  SUFFIX => '.tdf',
				  DIR => $self->{temp_dir});
    
    print $td_file <<EOF;
WLM {
    WLM_SYSTEM=SGE;
}

DATA {
EOF
    foreach (@transfer_files) {
	print $td_file "TRANSFER $_;\n";
    }
    print $td_file <<EOF;
}

EXECUTE {
EOF
    printf $td_file "DO %s ARGS=\"%s\" WLMARGS=\"%s\" PARMS=[1:%d;1];\n",
    $command, join (' ',@arguments), '-b y -@ '.$sge_options_file->filename, 
    $counter;
    print $td_file <<EOF;
}
EOF
    $td_file->close();

    # unfortunatly replic-submit does not synchronise with the jobs;
    # we have to gather the output, parse out the JOB ID and use SGE
    # to wait for the jobs
    my $rv = system (sprintf("/bin/sh -c \"cd %s; %s -e /dev/null -o /dev/null -sync y -b y %s %s %s %s\" 2>/dev/null >/dev/null",
			     $self->{temp_dir},
			     $qsub_binary,
			     (QSUB_OPTIONS || ""),
			     REPLICATOR_SUBMIT,
			     REPLICATOR_GATEWAY,
			     $td_file->filename));
    if ($rv == -1) {
	die "invoking replicator-submit failed: $!";
    }
    elsif ($rv != 0) {
	die "replicator-submit returned exit code $rv";
    }

    return @names;
}

sub wait_for {
    # replic-submit returns after the last job is finished, so we do
    # not need to wait for it actively
    return;
}

# iterate is called after submit, so all results should be there
sub iterate {
    my ($self, $success_callback, $error_callback) = @_;

    # since submit blocks until all jobs are ready, we can simply
    # iterate over all sequence names
    foreach my $seq (keys (%{$self->{job_info}})) {
	if ($self->get_status($seq) != JOB_FINISHED) {
	    &$error_callback($seq,'job finished with unclear conditions');
	}
	else {
	    &$success_callback($seq, $self->get_output($seq));
	}
	unlink $self->{job_info}->{$seq}->{output};
	unlink $self->{job_info}->{$seq}->{input};
    }
}
   
sub get_job_id {
    my ($self, $name) = @_;

    # simple return the name itself
    return $name;
}

# we cannot check whether a tool run was successfully, so
# return JOB_FINISHED in every case
sub get_status {
    my ($self, $name) = @_;

    # we currently do not have information about the job status
    return JOB_FINISHED;
}

sub get_output {
    my ($self, $name) = @_;

    my $handle = IO::File->new($self->{job_info}->{$name}->{output}, '<');
    return $handle;
}
