package Scheduler::Direct;
#
#  This file is part of GenDB 2.2
#  Copyright (C) 2005 CeBiTec, Bielefeld University
#                     <gendb@CeBiTec.Uni-Bielefeld.DE>
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#
=head1 NAME

Scheduler::Direct

=head1 DESCRIPTION

Special scheduler class for executing jobs on the local machine. This class acts as a fallback if no other scheduler classes are available or applicable.

=cut

    use strict;
use warnings;
use Scheduler qw(:JOB_FLAGS);
use FileUtils qw(get_tmp_file_object);
use base qw(Scheduler);

use IO::File;
use POSIX qw(WIFEXITED WEXITSTATUS WIFSIGNALED);

1;

# this scheduler is allways available
sub available {
    return 1;
}

sub new {
    my ($class, %parameters) = @_;
    my $self = $class->SUPER::new(%parameters);
    bless $self, ref $class || $class;
    return $self;
}

sub _substitute_string {
    my ($match, $subst) = @_;
    # escape meta characters in match and subst
    $match =~ s/\$/\\\$/g;
    $subst =~ s/\$/\\\$/g;
    return "s!$match!$subst!g";
}

# submit executes the jobs one by one
sub submit {
    my ($self) = @_;

    my @names = (keys %{$self->{sequences}}, keys %{$self->{data_sets}});
    my $counter = 0;
    foreach (@names) {
	$counter++;
	# create a temporary input file
	my $input_file = get_tmp_file_object($self->{temp_dir});
	my $output_file = $self->{temp_dir}."/$counter.stdout";
	$self->{job_info}->{$_}->{output} = $output_file;

	# fake job id
	$self->{job_info}->{$_}->{job_id} = $counter;

	# fill temp file
	if (exists($self->{sequences}->{$_})) {
	    print $input_file ">$_\n";
	    print $input_file $self->{sequences}->{$_};
	    print $input_file "\n";'>';
	}
	else {
	    print $input_file $self->{data_sets}->{$_};
	}
	$input_file->close();


	# substitute input placeholder in the commandline
	my ($command, @arguments) = @{$self->{commandline}};	
	@arguments = map { ($_ eq $self->{input_placeholder}) ? 
			       $input_file->filename : $_} @arguments;

	# the way we executed the commandline depend on the 
	# output processing. if a output placeholder was used,
	# we may easily redirect the output to the right file
	my $exit_code;
	if (defined ($self->{output_placeholder})) {
	    @arguments = map { ($_ eq $self->{output_placeholder}) ? 
				   $output_file : $_} @arguments;
	    # and execute the command...
	    $exit_code = system($command, @arguments);
	}
	else {
	    # if no output placeholder is set, the command is expected
	    # to write to stdout. we have to redirect stdout to the 
	    # output file
	    local *STDOUT;
	    open (OUTPUT, '>', $output_file) or die "unable to open file $output_file for writing";
	    *STDOUT = *OUTPUT;
	    $exit_code = system($command, @arguments);
	    close(OUTPUT);
	}
	# do some error checking....
	if ($exit_code == -1) {
	    # executing job itself has failed, print a log message
	    $self->log("executing '".join(" ",$command, @arguments)." has failed: $!");
	    $self->{job_info}->{$_}->{job_exit_code} = $exit_code;
	}
	elsif (WIFEXITED($exit_code)) {
	    $self->{job_info}->{$_}->{job_exit_code} = WEXITSTATUS($exit_code);
	}
	elsif (WIFSIGNALED($exit_code)) {
	    $self->log("executing '".join(" ",$command, @arguments)." has failed: process received signal ".WTERMSIG($exit_code));
	    $self->{job_info}->{$_}->{job_exit_code} = -1;
	}
	push @{$self->{job_ids}}, $counter;
	$self->{job_info}->{$_}->{job_id} = $counter;
    }
    return @{$self->{job_ids}} if (wantarray);
    return $self->{job_ids};
}

sub get_status {
    my ($self, $job_id) = @_;
    
    my @states;
    my @jobs = (defined ($job_id) ? ($job_id) : @{$self->{job_ids}});
    foreach (@jobs) {
	# find the right entry in our job management tables
	my $found = 0;
	foreach (values(%{$self->{job_info}})) {
	    if ($_->{job_id} == $_) {
		push @states, ($_->{job_exit_code} == 0) ?
		    JOB_FINISHED() : JOB_FAILED();
		$found = 1;
		last;
	    }
	}
	unless ($found) {
	    push @states, JOB_UNKNOWN;
	}
    }
    if (defined ($job_id)) {
	return $states[0];
    }
    return @states if (wantarray);
    return \@states;
}

# jobs are executed at submit time, so we don't need to wait for them
sub wait_for {
}

# jobs cannot be cancelled
sub cancel {
}

sub get_output {
    my ($self, $name) = @_;
    return undef unless ($self->{job_info}->{$name});
    unless (defined ($self->{job_info}->{$name}->{output})) {
	$self->log('undefined output file for sequence $name, not submitted yet ?');
	return undef;
    }
    my $handle = IO::File->new($self->{job_info}->{$name}->{output}, '<');
    return $handle;
}

sub get_job_id {
    my ($self, $name) = @_;
    return undef unless (defined ($self->{job_info}->{$name}));
    return $self->{job_info}->{$name}->{job_id};
}

sub can_cancel {
    return 0;
}
