package TCommon;

#
#  Copyright (C) 2007 CeBiTec, Bielefeld University
#  Written by Wolfgang Gerlach.
#  The original version (CARMA 1.2) was developed by Lutz Krause.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#

use strict;
use warnings;


use TCarma_phylo;
use TCarma_phylo_job;
use TCarma_MatePairFunctions;

use DBI;
use File::Temp qw(tempfile tempdir);

use TConfigs;


use Tree;

use Cwd;

my $DBH;
my $STH_GO;
my $STH_ACC;
my $STH_ID;

my $NCBI_NAMES;
my $NCBI_NODES;
my $CONSIDERED_RANKS_HASH;

1;


########################################
# Parase multiple fasta file
########################################
sub parse_fasta{
	my $file = shift @_;
	my $subref = shift @_;
	my @args = @_;
	
	
	unless(open (FASTA, "$file")){
		print STDERR "ERROR: parse_fasta: Could not open $file:$!\n"; 
		return;
	}
	

	my ($name,$sequence);
	
	# read each line in fasta file and process entrys
	# 
	while (<FASTA>) {
		chomp;
		next if /^\s*$/;
		next if /^\s*\#/;

		# check if line contains new fasta header
		if (/^>\s*(\S+)/) {
		
			# process fasta sequence
			if ($name) {
				&$subref($name,$sequence,@args) or return;
			}
			# new entry in fasta file read, therefore set sequence to empty string
			$name = $1;
			$sequence = "";
		}
		# line contains DNA sequence, append to $sequence
		else {
			$sequence .= $_;
		}
	}
	
	# submit job for remaining sequences
	if ($name) {
		&$subref($name,$sequence,@args) or return;
	}
	
	close (FASTA);
	
	
	return 1;
}


####################################################
# Parase multiple alignment in Stockholm format
####################################################
sub parse_stockholm{
	my $file = shift @_;
	my $subref = shift @_;
	my @args = @_;
	
	unless(open (IN, "$file")){
		print STDERR "ERROR: parse_fasta: Could not open $file:$!\n"; 
		return;
	}

	my $family_name;
	my $alignment = [];
	
	# read each line in fasta file and process entrys
	# 
	while (<IN>) {
		chomp;
		next if /^\s*$/;	# empty line

		if(/^\s*\#=GF\s+AC\s+(\S+)/){	#=GF <feature> <Generic per-File annotation, free text>
			$family_name = $1;
			$family_name =~ s/\.\d+//;
			next;
		}
	
		next if /^\s*\#/;
	
		# check for end of alignment
		if (/^\s*\/\/\s*$/) {
		
			# process alignment
			if ($family_name) {
				&$subref($family_name,$alignment,@args) or return;
			}
			# new entry in fasta file read, therfor set sequence to empty string
			$alignment = [];
			$family_name = undef;
		}
		# line contains sequence, add to alignment
		else {
			die "ERROR: wrong format: $_\n" unless /^\s*(\S+)\s+(\S+)\s*$/ ; #*
			my ($name,$sequence) = ($1,$2);
			push @$alignment, [$name,$sequence];
		}
	}
	
	# process last alignment
	if ($family_name) {
		&$subref($family_name,$alignment,@args) or return;
	}
	close (IN);

	return 1;
}



########################################
# Create matrix of pairwise distances
# for all sequences in a multiple alignment
########################################
sub create_matrix_OLD{
                                       #    my ($in,$format,$out,$dist_type,$out_acc,$phylip,$add_taxa,$verbose,$file) = @_;
my ($hash_of_precomputed_distances, $pfam_acc, $in,$format,$out,$dist_type,$out_acc,$phylip,$add_taxa,$verbose,$file) = @_;
    print STDERR "Creating matrix for file $in .... \n" if $verbose;

    my $fh_out;
    unless($phylip){
	open(OUT,">$out") or die $!;
	$fh_out = *OUT;
    }

    # get all sequences of multiple alignment
    my $s = parse_multiple_alignment($in,$format);
    
    return unless ref $s;
    
    my @sequences = @$s;
    
    return unless @sequences;

    my $seq_number = @sequences;

    #################################
    # Iterate over each sequence 
    # Replace gaps at beginning and end
    # of genomic fragments by '?'
    #################################
    for(my $j = 0; $j < $seq_number; $j++){
	my $acc = $sequences[$j]->[0];
	

	# check if sequence is a known member of pfam family
	next if _is_pfam($acc);

	my $seq = $sequences[$j]->[1];

	my $length = length($seq);
      
	if($seq =~ s/^([\.\-]+)//){
	
	    my $gap_leading = length($1);
    
	    my $prefix = '?' x $gap_leading;
	    $seq = "${prefix}${seq}";	    
	}
	
	if($seq =~ s/([\.\-]+)$//){
	    my $gap = length($1);
	    
	    my $suffix = '?' x $gap;
	    $seq = "${seq}$suffix";
	}

	die "ERROR: Wrong length ($length) ".length($seq)."\n" unless(length($seq) == $length);

	$sequences[$j] = [$acc,$seq];
    }


    
    my $c = 0;
    
    my %dist_to_read = ();
    my @missing_dist = ();
    my @reads = ();
    
    ################################
    # Iterate over each sequence
    ################################
    for(my $i = 0; $i < $seq_number; $i++){
	my $si = $sequences[$i];
	
	my $acc1 = $si->[0];
	
	my $is_read_1 = 1;

	# check if sequence is a known pfam member
	if (TCommon::_is_pfam($acc1)){		    
	    $is_read_1 = 0;
	}
	else{
	    push @reads,$acc1;
	}
	
	my $l1 = length($si->[1]);
	
	$c++;
	print STDERR "\r$c/$seq_number" if $verbose;
	
	unless($phylip){
	    print $fh_out "$acc1,$acc1=0\n"; 
	}

	############################
	# Iterate over sequences
	# and caclulate pairwise distances
	############################
	for(my $j = $i+1; $j< $seq_number; $j++){	    
	    my $sj = $sequences[$j];
	    
	    my $acc2 = $sj->[0];
	    
	    my $l2 = length($sj->[1]);
	    
	    die unless $l2 == $l1;
	    
	    # Get pairwise distance
	    my $dist = _dist(\$si->[1],$acc1,\$sj->[1],$acc2,$l1,$dist_type,$file);

	    die "ERROR: $dist\n" if $dist and $dist < 0;
	    die "ERROR: $dist" if $dist and $dist > 1;

	    if(defined $dist){
		die if exists $dist_to_read{$acc2}->{$acc1};
		die "ERROR 01: $acc1 $acc2\n" if exists $dist_to_read{$acc1}->{$acc2};
		
		$dist_to_read{$acc1}->{$acc2} = $dist;
		
		unless($phylip){
		    print $fh_out "$acc1,$acc2=$dist\n";
		}
	    }
	    else{
		push @missing_dist, [$acc1,$acc2];
	    }		    		    
	}
    }
    
    ####################################
    # Get missing disances, i.e. pairwise
    # distance of sequences without
    # sufficient sequence overlap
    #####################################
    foreach my $pair (@missing_dist){
	my ($acc1,$acc2) = @$pair;

	my $missing_value = _get_missing_value($acc1,$acc2,\%dist_to_read);
	die unless defined $missing_value;
	
	die "ERROR: $missing_value\n" if $missing_value < 0;
	die "ERROR: $missing_value" if $missing_value > 1;
	
	unless($phylip){
	    print $fh_out "$acc1,$acc2=$missing_value\n";
	}
	else{
	    $dist_to_read{$acc1}->{$acc2} = $missing_value;
	}
    }

    ############################
    # Print matrix
    ############################
    unless(open(OUT,">$out")){
	print STDERR "ERROR: Could not open file $out:\n$!\n";
	return;
    }
    
    unless(open(ACC,">$out_acc")){
	print STDERR "ERROR: Could not open file $out_acc:\n$!\n";
	return;
	}

	$fh_out = *OUT;	

	my @acc = keys %dist_to_read;
	my $acc_number = @acc;
	print $fh_out "\t$acc_number\n";

	my $id = 0;
	foreach my $acc1 (@acc){
	    $id++;

	    if($id > 999999999){
		die;
	    }
	    my $length_id = length("$id");
	    
	    my $id_str = $id;
	    
	    my $diff = 12 - $length_id;
	    
	    $id_str .= (' ' x $diff);
	    
	    print $fh_out $id_str;
	    
	    my $tmp_acc = $acc1;
	    
	    $tmp_acc =~ s/^\d+,//;

	    print ACC "$id\t$tmp_acc\n";
	    
	    foreach my $acc2 (@acc){
		my $d;
		
		if($acc1 eq $acc2){
		    $d = '0.0';
		}
		else{
		    # just a check ...
		    die "ERROR 12: $acc1 $acc2\n" if (exists $dist_to_read{$acc1}->{$acc2}) and 
			(exists $dist_to_read{$acc2}->{$acc1});

		    $d = exists $dist_to_read{$acc1}->{$acc2} ? 
			$dist_to_read{$acc1}->{$acc2} : 
			    $dist_to_read{$acc2}->{$acc1};
		}
		die "ERROR no dist for: $acc1 $acc2\n" 
		    unless defined $d;
		print $fh_out "$d ";
	    }
	    print $fh_out "\n";		    
	}		
    	    
    
    
    close($fh_out);
    
    close(ACC);
    
    print STDERR "\nDone.\n" if $verbose;
    
    return 1;
}

########################################
# Create matrix of pairwise distances
# for all sequences in a multiple alignment
########################################
sub create_matrix{
	my ($hash_of_precomputed_distances, $pfam_acc, $in,$format,$out,$dist_type,$out_acc,$phylip,$add_taxa,$verbose) = @_;
	# in = alignment (hmm alignment output)
	# out = matrix_tmp_file

	print STDERR "Creating matrix for file $in .... \n" if $verbose;

	my $fh_out;
	unless($phylip){
		open(OUT,">$out") or die $!;
		$fh_out = *OUT;
	}
	
	unless (-e $in) {
		print STDERR "ERROR (create_matrix): File $in does not exist!\n";
		die;
	}
	
	# get all sequences of multiple alignment
	
	my $s = parse_multiple_alignment($in,$format);

	return unless ref $s;
	my @sequences = @$s;
	return unless @sequences;

	


	my $seq_number = @sequences;

	#################################
	# Iterate over each sequence 
	# Replace gaps at beginning and end
	# of genomic fragments by '?'
	#################################
	for(my $j = 0; $j < $seq_number; $j++){
		my $acc = $sequences[$j]->[0];
	

		# check if sequence is a known member of pfam family
		next if _is_pfam($acc);

		my $seq = $sequences[$j]->[1];

		my $length = length($seq);

		if($seq =~ s/^([\.\-]+)//){
	
			my $gap_leading = length($1);
	
			my $prefix = '?' x $gap_leading;
			$seq = "${prefix}${seq}";		
		}
	
		if($seq =~ s/([\.\-]+)$//){
			my $gap = length($1);
		
			my $suffix = '?' x $gap;
			$seq = "${seq}$suffix";
		}

		

		die "ERROR: Wrong length ($length) ".length($seq)."\n" unless(length($seq) == $length);

		$sequences[$j] = [$acc,$seq];
	}


	#if ($main::opt_m) {   #}
	#		TCarma_MatePairFunctions::mergeMatePairs(\@sequences);
	#}
	$seq_number = @sequences;
	
	my %dist_to_read = ();
	my %dist_to_read_precomputed;
	my $precomputed_distances=1;
	
	unless (exists $hash_of_precomputed_distances->{$pfam_acc}) {
		$hash_of_precomputed_distances->{$pfam_acc} = {};
		$precomputed_distances=0;
	}

	
	
	
	my $test = $hash_of_precomputed_distances->{$pfam_acc};
	%dist_to_read_precomputed = %$test;

	
	my $acc1;
	my $acc2;
	my $dist;
	my $tmp_acc1;
	my $tmp_acc2;
	my $acc1_is_pfam=0;
	my $acc2_is_pfam=0;
	
	
	
	my $precomputed_distances_file='';

	if (DIST_CACHE_DIR) {
		$precomputed_distances_file= DIST_CACHE_DIR . "\/".$pfam_acc.".matrix";  
		$precomputed_distances_file=~s/\/\//\//;
	}

	
	if ((-e $precomputed_distances_file) && (keys (%dist_to_read_precomputed) == 0) ) {
		#######################
		# read cached distances from file
		#######################

		$precomputed_distances=1;
		open(PRE_DIST_READ,"<$precomputed_distances_file") or die $!;

		my $line = <PRE_DIST_READ>;
		chomp($line);
		my @pfam_accs = split(/\t/, $line);
		my $outer_acc_position = 0;

		while ($line = <PRE_DIST_READ>) {

			my $inner_acc_position = 0;
			$acc1 = $pfam_accs[$outer_acc_position];

			chomp($line);
			my @dists = split(/ /, $line);	
			
			foreach $dist (@dists) {
				$acc2 = $pfam_accs[$outer_acc_position+$inner_acc_position+1];

				
				unless (defined $acc2) {
					print STDERR "outer_acc_position: $outer_acc_position\n";
					print STDERR "inner_acc_position: $inner_acc_position\n";
					print STDERR "dists :".@dists."\n";
					print STDERR "pfam_accs: ".@pfam_accs."\n";
					print STDERR "ERROR: acc2 undefined.\n";
					die;
				}

				if  (   (exists $dist_to_read_precomputed{$acc1}->{$acc2}) ||
					(exists $dist_to_read_precomputed{$acc2}->{$acc1}) ) {
					print STDERR "ERROR: already exists!?\n";
					print STDERR "ERROR: pfamacc: $pfam_acc, acc1: $acc1, acc2: $acc2\n";
					}				
				
				$dist_to_read_precomputed{$acc1}->{$acc2} = $dist;

				if  (   (exists $dist_to_read_precomputed{$acc1}->{$acc2}) &&
					(exists $dist_to_read_precomputed{$acc2}->{$acc1}) ) {
					print STDERR "ERROR: both exist!? $acc1, $acc2\n";
					
					}				
							
				
				
					
				$inner_acc_position++;
			}
			#die;
			$outer_acc_position++;
		}
		
		close(PRE_DIST_READ);
		
		print STDERR "done.\n" if ($main::opt_v);
	}
	
	
	
	my @missing_dist = ();
	#my @reads = ();
	#my $distance_already_contained;

	################################
	# calculate pairwise distances
	################################
	for(my $i = 0; $i < $seq_number; $i++){
#	print "i: $i\n";
		my $acc_seq_pair = $sequences[$i];
	
		$acc1 = $acc_seq_pair->[0];
	
		my $acc1_is_pfam;

		# check if sequence is a known pfam member
		if (TCommon::_is_pfam($acc1)){			
			$acc1_is_pfam = 1;
		} else {
			$acc1_is_pfam = 0;
		}
		#else{
			#push @reads,$acc1;
		#}
	
		my $l1 = length($acc_seq_pair->[1]);
#print STDERR "1) ".$acc_seq_pair->[1]."\n";

		unless($phylip){
			print $fh_out "$acc1,$acc1=0\n"; 
		}

		$tmp_acc1 = $acc1;
		$tmp_acc1 =~ s/^\d+,//; # delete number in front of description line		
		
		for(my $j = $i+1; $j< $seq_number; $j++){		
			my $sj = $sequences[$j];
			
			$acc2 = $sj->[0];

#print STDERR "2) ".$sj->[1]."\n";			
			$tmp_acc2 = $acc2;
			$tmp_acc2 =~ s/^\d+,//; # delete number in front of description line		
			$acc2_is_pfam = 0;
			if (TCommon::_is_pfam($acc2)){			
				$acc2_is_pfam = 1;
			}
			
			if ($precomputed_distances && $acc1_is_pfam && $acc2_is_pfam)  {
				unless($phylip){
					$dist = exists $dist_to_read_precomputed{$tmp_acc1}->{$tmp_acc2}?
							$dist_to_read_precomputed{$tmp_acc1}->{$tmp_acc2}:
							$dist_to_read_precomputed{$tmp_acc2}->{$tmp_acc1};

					print $fh_out "$acc1,$acc2=$dist\n";
				}
				next; # already stored, skip!
			}
		
			my $l2 = length($sj->[1]);

			die unless $l2 == $l1;
		
			undef $dist;
			
			$dist = _dist(\$acc_seq_pair->[1],$acc1,\$sj->[1],$acc2,$l1,$dist_type);
			

			if(defined $dist){
			
				die "ERROR: $dist\n" if $dist and $dist < 0;
				die "ERROR: $dist" if $dist and $dist > 1;
				if ( exists $dist_to_read{$acc2}->{$acc1} && exists $dist_to_read{$acc1}->{$acc2}) {
					print STDERR "ERROR: both exist!!\n";
					print STDERR "ERROR, dist: $dist\n";
					print STDERR "ERROR, reverse: ".$dist_to_read{$acc2}->{$acc1}."\n";
					die;
				}
				if ( exists $dist_to_read{$acc1}->{$acc2} && $precomputed_distances==0 ) {
					print STDERR "ERROR 01: $acc1 $acc2 already hashed!?\n" ;
					die;
				}

				if ($precomputed_distances) {
					$dist_to_read{$acc1}->{$acc2} = $dist; # pfam-pfam entries already skipped...
				} else {
					if ($acc1_is_pfam && $acc2_is_pfam) {  
						$dist_to_read_precomputed{$tmp_acc1}->{$tmp_acc2} = $dist;
					} else {# if EGT is contained:
						$dist_to_read{$acc1}->{$acc2} = $dist;
					}				
				}
				
				
				if ( exists $dist_to_read{$acc2}->{$acc1} && exists $dist_to_read{$acc1}->{$acc2}) {
					print STDERR "ERROR2: both exist!!:  $acc1 $acc2\n";
					print STDERR "ERROR2, dist: $dist\n";
					print STDERR "ERROR2, reverse: ".$dist_to_read{$acc2}->{$acc1}."\n";
					die;
				}
		
				unless($phylip){
					print $fh_out "$acc1,$acc2=$dist\n";
				}
			}
			else{
				push @missing_dist, [$acc1,$acc2];
			}
					
		}
	}
	
	


	####################################
	# Get missing disances, i.e. pairwise
	# distance of sequences without
	# sufficient sequence overlap
	#####################################
	foreach my $pair (@missing_dist){
		my ($acc1,$acc2) = @$pair;

		my $missing_value = _get_missing_value($acc1,$acc2,\%dist_to_read);
		unless (defined $missing_value) {
			print STDERR "missing_value not defined!";
			die;
		}
	
		die "ERROR: $missing_value\n" if $missing_value < 0;
		die "ERROR: $missing_value" if $missing_value > 1;
	
		unless($phylip){
			print $fh_out "$acc1,$acc2=$missing_value\n";
		}
		else{

			if ($precomputed_distances) {
				$dist_to_read{$acc1}->{$acc2} = $missing_value; # pfam-pfam entries should not be missing
			} else {
				if (TCommon::_is_pfam($acc1) && TCommon::_is_pfam($acc2)) {  
					$tmp_acc1 = $acc1;
					$tmp_acc1 =~ s/^\d+,//;
					$tmp_acc2 = $acc2;
					$tmp_acc2 =~ s/^\d+,//;
				
					$dist_to_read_precomputed{$tmp_acc1}->{$tmp_acc2} = $missing_value;
				} else {# if EGT is contained:
					$dist_to_read{$acc1}->{$acc2} = $missing_value;
				}				
			}
		}
	}

	############################
	# Print matrix
	############################
	unless(open(OUT,">$out")){
		print STDERR "ERROR: Could not open file $out:\n$!\n";
		return;
	}
	
	unless(open(ACC,">$out_acc")){
		print STDERR "ERROR: Could not open file $out_acc:\n$!\n";
		return;
	}

	
	$fh_out = *OUT;	

	my @acc = keys %dist_to_read;

	if ($precomputed_distances==0){	
		open(PRE_DIST_WRITE,">$precomputed_distances_file.part") or die $!;
		
		#write accs:
		foreach my $acc1 (@acc) {
			if (TCommon::_is_pfam($acc1) ){
				$tmp_acc1 = $acc1;
				
				$tmp_acc1 =~ s/^\d+,//; # delete number in front of description line
				print PRE_DIST_WRITE $tmp_acc1."\t";
			}
		}
		print PRE_DIST_WRITE "\n";	
	}
	
	my $acc_number = @acc;
	print $fh_out "\t$acc_number\n";
	my $num_pfam_seqs=0; 

	
	my $newline=1;

	my $id = 0;
	
	
	
	foreach my $acc1 (@acc){

		$id++;

		if($id > 999999999){
			die;
		}
		
		
		$acc1_is_pfam=TCommon::_is_pfam($acc1);
		
		if ($precomputed_distances==0){			
			if ($acc1_is_pfam) {
				$num_pfam_seqs++;
				if ($num_pfam_seqs>1) {
					print PRE_DIST_WRITE "\n";
					$newline=1;
				}
			}
		}
		
		
		my $length_id = length("$id");
		my $id_str = $id;
		my $diff = 12 - $length_id;
		$id_str .= (' ' x $diff);
		
		print $fh_out $id_str;
		
		$tmp_acc1 = $acc1;
		$tmp_acc1 =~ s/^\d+,//; # delete number in front of description line

		print ACC "$id\t$tmp_acc1\n";
		my $j = 0;
		foreach my $acc2 (@acc){
			my $d;
			$j++;

			$tmp_acc2 = $acc2;
			$tmp_acc2 =~ s/^\d+,//; # delete number in front of description line
		

			$acc2_is_pfam=TCommon::_is_pfam($acc2);
			
		
		
			if($acc1 eq $acc2){
				$d = '0.0';
			} elsif ($acc1_is_pfam && $acc2_is_pfam) {
				$d = exists $dist_to_read_precomputed{$tmp_acc1}->{$tmp_acc2} ? 
						$dist_to_read_precomputed{$tmp_acc1}->{$tmp_acc2} : 
						$dist_to_read_precomputed{$tmp_acc2}->{$tmp_acc1};
				unless (defined $d) {
					print STDERR "ERROR($pfam_acc) no dist for: $tmp_acc1 $tmp_acc2\n";
					die;
				}		
			}
			else{
				# just a check ...
				if ( (exists $dist_to_read{$acc1}->{$acc2}) and (exists $dist_to_read{$acc2}->{$acc1}) ) {
					print STDERR "ERROR: both exist: $acc1 $acc2\n";
					print STDERR "ERROR: ".$dist_to_read{$acc1}->{$acc2}.",".$dist_to_read{$acc2}->{$acc1}."\n";
					die;
				}
				$d = exists $dist_to_read{$acc1}->{$acc2} ? 
					$dist_to_read{$acc1}->{$acc2} : 
					$dist_to_read{$acc2}->{$acc1};
				unless (defined $d) {
					print STDERR "ERROR($pfam_acc) no dist for: $acc1 $acc2\n";
					die;
				}
	
			}
			
			print $fh_out "$d ";
			
			
			if ($precomputed_distances==0 && $acc1_is_pfam && $acc2_is_pfam && ($id<$j)){
				
				if ($newline==0) {
					print PRE_DIST_WRITE " ";
					
				} else {
					$newline=0;
				}
				print PRE_DIST_WRITE substr($d,0,8); # length defines accuracy
					
			}
			
		}
		print $fh_out "\n";			
	}		
	
	close($fh_out);
	
	close(ACC);
	
	if ($precomputed_distances==0) {
		close(PRE_DIST_WRITE);
		# to make sure the matrix file is really complete:
		rename("$precomputed_distances_file.part",$precomputed_distances_file) || die "ERROR (rename): $!";
	}	
	
	print STDERR "\nDone.\n" if $verbose;
	
	
	return 1;
}

#####################################################
# Check if sequence is a known member of Pfam family
######################################################
sub _is_pfam{
	my ($acc) = @_;

	if ($acc =~ /^\d*,?c_/ ) {
		return 0;
	}

	return 1;
}

###########################
# Reconstruct phylogenetic tree
# using the neighbor program
# from the Phylip package
###########################
sub run_phylip_nj{
	my ($matrix,$tmp_dir,$verbose) = @_;

	my $tree_file = "$tmp_dir/outtree";

	die if -e $tree_file;
	
	my ($fh, $tmp_file) = tempfile(UNLINK => $main::opt_t?0:1, DIR => $tmp_dir, SUFFIX => '.njtree');
	print $fh "$matrix\nY\n";
	close($fh);
	
	my $cdir = getcwd;

	chdir($tmp_dir);

	my $cmd = PHYLIP_NJ_EXEC . " < $tmp_file > /dev/null";

	
	unless(system($cmd) == 0){
		print STDERR "ERROR: (run_phylip_nj) Could not run $cmd!\n";
		print STDERR "$!\n";
		return;
	}
	print STDERR "Done.\n" if $verbose;

	chdir($cdir);

	unlink($tmp_file) unless ($main::opt_t);

	return $tree_file;
}

#################################
# Predict taxonomic origin of
# DNA fragments based on their
# location in phylogenetic tree
#################################
sub predict_taxa{
	my ($tree_file,$acc_by_id_file,$verbose,$alignment_file) = @_;

	print STDERR "Parsing tree ... " if $verbose;

	my ($tree,$read_nodes) = Tree::get_tree($tree_file,$acc_by_id_file);
	print STDERR "Done. ".@$read_nodes." reads.\n" if $verbose;

	my @predictions = ();
	
	print STDERR "Predicting taxa ... " if $verbose;
	foreach my $read (@$read_nodes){
	
		my $taxa =Tree::get_fragment_taxa($read,$alignment_file);
		#print "taxa: " . $taxa . "\n";
		$taxa = "0" unless $taxa;

		my $acc = $read->{a};
	
		push @predictions,[$acc,$taxa];

	}
	print STDERR " Done.\n" if $verbose;

	return \@predictions;
}

###############################
# Get sequence identity of
# two aligned sequence
###############################
sub _dist{
	my ($seq1,$acc1,$seq2,$acc2,$length,$type) = @_;

	my $ident = 0;
	my $len = 0;

	my $counter = 0;
	
	my $score;	
	
	my $s1 = $$seq1;
	my $s2 = $$seq2;

	my $missing = 0;

	while($s1 =~ /(.)/g){
		$counter++;
	
		my  $aa1 = $1;
	
		$aa1 = '?' if $aa1 eq 'X';
		$aa1 = '?' if $aa1 eq 'Z';

		die "\nERRROR: counter: $counter\n$s2\n" unless $s2 =~ /(.)/g;

		my $aa2 = $1;

		$aa2 = '?' if $aa2 eq 'X';
		$aa2 = '?' if $aa2 eq 'Z';
		
		if( ($aa1 eq '?') or ($aa2 eq '?') ){
			$missing++;
			next;
		}
		elsif(($aa1 eq '-') or ($aa2 eq '-')){
			next if $aa1 eq $aa2;
			$len += 1; 
			next;
		}
		elsif($aa1 eq $aa2){  # TODO PAM/blosum or similar!
			$len++;
			$ident++;
			next;
		}
		else{
			$len++;
			next;
		}
	}
		
	# Check if overlapping region is sufficient, i.e. more than 10 aa
	if(($len < 10) and ($len < $length) ){

		# both sequences are known family members
		if(_is_pfam($acc1) and _is_pfam($acc2)){
			# Set distance to 1 if overlap is less than 3 aa
			return 1 if $len < 3;
			# Distance cannot be determined if overlap is less than 6 aa
			return if $len < 6;
		}

		# both sequences are environmental genomic fragments
		# distance cannot be determined
		elsif( (! _is_pfam($acc1) ) and (! _is_pfam($acc2) ) ){
			return;
		}

		# one environmental fragment, one known member
		else{
			# Set distance to 1 if overlapping region is shorter than 3 aa
			return 1 if $len < 3;
			# Otherwise, distance cannot be determined
			return unless ($len > 5 and $missing < 5);
		}
	}
	
	# convert sequence similarity to distance
	# and return it
	return (1 - $ident/$len);
}


##########################################
# Estimate distance of two items using
# the additive estimation as proposed by
# Landry et al.
##########################################
sub _get_missing_value{
	my ($acci,$accj,$dist) = @_;

	my ($acc_ref,$dist_to_ref);

	my @shared = ();
	while ( ($acc_ref,$dist_to_ref) = each %{ $dist->{$acci} }){
		if(exists $dist->{$accj}->{$acc_ref}){
			push @shared,$acc_ref;
		}
	}

	my ($l,$k) = (0,0);
	my $minmax = '1.0';

	my $shared_number = @shared;
	
	while($l < $shared_number){
		my $accl = $shared[$l];
	
		$k = $l + 1;
	
		while($k < $shared_number){
			my $acck = $shared[$k];

			my $dist_k_l;
			if (exists $dist->{$acck}->{$accl}){
				$dist_k_l = $dist->{$acck}->{$accl} 
			}
			elsif(exists $dist->{$acck}->{$accl}){
				$dist_k_l = $dist->{$accl}->{$acck};
			}
			unless(defined $dist_k_l){
				$k++;
				next;
			}

			die "\nERROR: no dist $acck $accl\n" unless defined $dist_k_l;

			my $max = max( ($dist->{$acci}->{$acck} + $dist->{$accj}->{$accl}), 
						 ($dist->{$acci}->{$accl} + $dist->{$accj}->{$acck}) ) - $dist_k_l;
		
		
			if($max < 0){
				return 0;
			}


			unless(defined $minmax){
				$minmax = $max;
			}
			elsif($max < $minmax){
				$minmax = $max;
			}

			$k++;
		}

		$l++;
	}
	die unless defined $minmax;
	return $minmax;
}

########################################
# Parse multiple alignment in nexus
# or stockholm format
########################################
sub parse_multiple_alignment{
	my ($alignment,$format,$add_taxa) = @_;

	unless(open(AL,"<$alignment")){
		print STDERR "ERROR （parse_multiple_alignment）: Could not open file $alignment:\n$!\n";
		exit 1;
	}
	#my $sep = $TConfigs::SEPSYMBOL;
	my $sep = SEPSYMBOL;
	
	my @sequences = ();
	print STDERR "format: $format\n" if ($main::opt_v);

	##############################################
	# Parse multiple alignment in nexus format
	##############################################
	if($format eq 'nexus'){
		my $line;
		my $matrix = 0;
		my $counter = 0;
	
		while ($line = <AL>){
			next if $line =~ /^\s*$/;
		
			if($matrix){
				last if $line =~ /;/;
				chomp($line);
		
				die "ERROR: wrong format:\n$line\n" unless $line =~ /^(\S+)\s+(\S+)$/;
		
				my ($acc,$seq) = ($1,$2);

				if($add_taxa){
					my $tmp_acc = $acc;
					$tmp_acc =~ s/^c_//;
					
					my $tax = get_taxa($tmp_acc);
					
					die unless @$tax;
			
					
					$tax = join($sep,@$tax);
			
					$tax =~ s/\(class\)$sep//g;
			
					if(length($tax) > 20){
						$tax = substr($tax,0,20);
						$tax =~ s/$sep$//;
					}
			
					if($acc =~ /c$sep/){
						$acc = "c".$sep.$tax;
					}
					else{
						$acc = $tax;
					}
				}		
		
				$counter++;
		
				push @sequences, ["$counter$acc",uc($seq)];		
			}
			else{
				if($line =~ /^matrix/){
					$matrix = 1;
				}
			}
		}
		unless($matrix){
			print STDERR "ERROR: alignment file $alignment has wrong format\n";
			return;
		}
	}

	##############################################
	# Parse multiple alignment in stockholm format
	##############################################
	elsif($format eq 'stockholm'){
		die "ERROR: not implemented yet\n" if $add_taxa;
	
		my $line;
		my %seqs = ();

		while($line = <AL>){
		#print "read: $line \n";
			last if $line =~ /^\#=GF AU    HMMER 2.3.2/;
		}
		while($line = <AL>){
		#print "read: $line \n";
			next if $line =~ /^\s*$/;
			next if $line =~ /^\#=GC RF/;
			last if $line =~ /^\/\//;
			die "ERROR: wrong line format in file $alignment\n($line)\n" unless $line =~ /^(\S+)\s+(\S+)\s*$/;
			my ($acc,$seq) = ($1,$2);

			if(exists $seqs{$acc}){
				$seqs{$acc} .= $seq;
			}
			else{
			
				$seqs{$acc} = $seq;
			}
		}

		#print "seqs: ${%seqs} \n";
		my $counter = 0;
		foreach my $acc (keys %seqs){
			$counter++;
			my $seq = $seqs{$acc};
			delete $seqs{$acc};
		
			push @sequences, ["${counter},$acc",uc($seq)];	# Without the comma after the counter it was a bug!!	
		}
	}
	else{
		print STDERR "ERROR: Unknown file format $format!\nSupported formats: nexus, stockholm\n";
		die;
	}
	close(AL);


	unless(@sequences){
		print STDERR "ERROR (parse_multiple_alignment): no sequences in file $alignment\n";
		return;
	}	
	
	return \@sequences;
}


##############################
# Get the maximum of two numbers
##############################
sub max{
	my ($a,$b) = @_;

	return $a if $a >= $b;

	return $b;
}


##############################
# Initialize database handle
# to Pfam MySQL database
##############################
sub init_dbh{
	$DBH = DBI->connect("dbi:mysql:database=".PFAM_MYSQL_DATABASE.
			";host=". DBHOST,MYSQL_USER,MYSQL_PASSWD);
	
	die "ERROR: ".$DBI::errstr unless ref $DBH;

	return 1;
}

sub check_pfam_version{

	unless (defined PFAM_VERSION) {
		die "ERROR: PFAM_VERSION is not defined!\n";
	}

	if (PFAM_VERSION eq "") {
		die "ERROR: PFAM_VERSION is empty!\n";
	}

	if (PFAM_VERSION eq "-1") {
		return 1;
	}

	if (PFAM_VERSION eq "0") {
		return 1;
	}

	unless(ref $DBH){
		die unless init_dbh();
	}

	my $st="select pfam_release from VERSION;";
	
	my $array_ref = $DBH->selectall_arrayref($st);

	my $it=0;
	while (not ref $array_ref) {
		print STDERR "ERROR: get version number: \"$!\", st:\"$st\"\n";
		init_dbh();
		if ($it > 2) {
			
			die "Enough tries. Doesn't work. :(\n";
		}
		print STDERR "Try again to contact MySQL server. $it\n";
		sleep(2);
		$array_ref = $DBH->selectall_arrayref($st);
		if (ref $array_ref) {
			print STDERR "Success contacting MySQL server. $it\n";
		}
		$it++;
	}
	
	my $version = $array_ref->[0]->[0];

	unless (defined $version) {
		print STDERR "ERROR: Could not read the Pfam version from the mysql-table !\n";
		die;	
	}

	if (PFAM_VERSION eq $version) {
		print STDERR "PFAM_VERSION=$version\n";
	} else {
		print STDERR "ERROR: Expected Pfam version ".PFAM_VERSION." but the Pfam mysql-tables are \"$version\" !\n";
		die;
	}


	return 1;
}



sub get_taxa{ 
	my ($acc) = @_;

	unless(ref $DBH){
		die unless init_dbh();
	}

	$acc =~ s/\/\d+-\d+\s*$//;

	my $st = "select ncbi_code from pfamseq where pfamseq_id = \"$acc\"";
	

	my $array_ref = $DBH->selectall_arrayref($st);
   	my $it=1;
	while (not ref $array_ref) {
		print STDERR "ERROR: get taxa: \"$!\", st:\"$st\"\n";
		init_dbh();
		if ($it > 5) {
			die "Enough tries. Doesn't work. :(\n";
		}
		print STDERR "Try again to contact MySQL server. $it\n";
		sleep(2);
		$array_ref = $DBH->selectall_arrayref($st);
		if (ref $array_ref) {
			print STDERR "Success contacting MySQL server. $it\n";
		}
		$it++;
	}

	unless(@$array_ref == 1){
		print STDERR "ERROR: tax @$array_ref \n $st\n" if @$array_ref > 1;
		return ["0"];
	}

	my $ncbi_code = $array_ref->[0]->[0];
	
	unless (defined $ncbi_code) {
			die;
	}
		
	my @ret=();

	push(@ret, $ncbi_code);

	unless ($NCBI_NODES) {
		init_NCBI_NODES();
	}

	while ($ncbi_code > 1) { 

		unless (defined @{$NCBI_NODES->{$ncbi_code}} ) {
			print STDERR "ncbi_code $ncbi_code in NCBI_NODES not found !?\n";
			#exit(1);
			return ["0"];
		}
		
		($ncbi_code) = @{$NCBI_NODES->{$ncbi_code}};
		
		unless ($ncbi_code) {
			die;
		}
				
		if ($ncbi_code > 1) {
	
			unshift(@ret, $ncbi_code);
		}

	}
	
		
	return \@ret;
}


sub init_NCBI_NODES(){
	open (STREAM, NODES_FILE);
	
	while (my $line = <STREAM>) {
		#print $line;
		my ($ncbi_id, $parent_id, $rank);
		unless ( ($ncbi_id, $parent_id, $rank)=split(/\s*\|\s*/, $line) ) {
			die;
		}
		#print "(".$ncbi_id.",".$parent_id.",".$rank.")\n";
		$NCBI_NODES->{$ncbi_id}=[$parent_id, $rank];
	}

	close(STREAM);


	open (STREAM2, MERGED_FILE);

	while (my $line = <STREAM2>) {
		#print $line;
		my ($old_id, $new_id);
		unless ( ($old_id, $new_id) =split(/\s*\|\s*/, $line) ) {
			die;
		}
		
		if (defined $NCBI_NODES->{$new_id} ) {
		
			unless (defined $NCBI_NODES->{$old_id} ) {
			
				$NCBI_NODES->{$old_id}=$NCBI_NODES->{$new_id};
			}
		} 
		
		
		
	}
	
	
	close(STREAM2);	


	unless ($NCBI_NODES) {
		die;
	}
	
}


sub init_NCBI_NAMES(){
	open (STREAM, NAMES_FILE);
	
	while (my $line = <STREAM>) {
		#print $line;
		my ($ncbi_id, $name, $nametype);
		unless (($ncbi_id, $name, undef, $nametype) =split(/\s*\|\s*/, $line) ) {
			die;
		}
		#print "GOT: [$ncbi_id, $name, $nametype]\n";
		if ($nametype eq "scientific name") {
			$NCBI_NAMES->{$ncbi_id}=$name;	
			#print $name."\n";
		}
	}

	close(STREAM);

	unless ($NCBI_NAMES) {
		die;
	}
	
	# include MERGED_FILE
	
	
	open (STREAM2, MERGED_FILE);

	while (my $line = <STREAM2>) {
		#print "LINE!!!! ".$line;
		my ($old_id, $new_id);
		unless (($old_id, $new_id) =split(/\s*\|\s*/, $line) ) {
			die;
		}
		#print "LINE: $line";
		#print "mapping: $old_id, $new_id\n";
		if (defined $NCBI_NAMES->{$new_id} ) {
				#print STDERR "new $new_id exists\n";		
			unless (defined $NCBI_NAMES->{$old_id} ) {
			
				$NCBI_NAMES->{$old_id}=$NCBI_NAMES->{$new_id};
			}
		} 
		
		
		
	}
	
	
	close(STREAM2);	
	
}

sub init_CONSIDERED_RANKS_HASH {
	
	my %hash= map { lc($_) } CONSIDERED_RANKS;

	unless (%hash) {
		die;
		}

	$CONSIDERED_RANKS_HASH=\%hash;

}

############################
# gives the taxon for a certain ncbi_code
# example: cellular organisms!Bacteria!Actinobacteria!Actinobacteria (class)!Actinobacteridae!Actinomycetales
############################
sub get_taxon_for_ncbi_code{
	my ($ncbi_code) = @_;

	if ($ncbi_code==0) {
		return "Unknown";
	}

	unless ($NCBI_NAMES) {
		init_NCBI_NAMES();
	}
	
	unless ($NCBI_NODES) {
		init_NCBI_NODES();
	}
	
	if (CONSIDERED_RANKS_MODUS eq 'none') {
		print STDERR "ERROR: Should not have called this fundtion.\n";
		return "";
	}

	unless ($CONSIDERED_RANKS_HASH)  {
		init_CONSIDERED_RANKS_HASH();	
	}

	my $rank;
	my $name;
	my $parent;
	my $taxon="";
	
	

	
	
	while ($ncbi_code!=1) {
		
		unless (defined @{$NCBI_NODES->{$ncbi_code}}) {
			print STDERR "Got nothing for ncbi_code in ncbi_nodes: $ncbi_code\n";
			exit(1);
		}
				
		unless ( ($parent, $rank) = @{$NCBI_NODES->{$ncbi_code}} ) {
			die;
		}
		
		#print "rank: $rank\n";
		if ((lc(CONSIDERED_RANKS_MODUS) eq 'all') || (defined $CONSIDERED_RANKS_HASH->{lc($rank)} && $CONSIDERED_RANKS_HASH->{lc($rank)}==1 )) {

			$name = $NCBI_NAMES->{$ncbi_code};
		#print "true name: $name\n";
			unless ($name) {
				die;
				}
		
			if ($taxon) {
				$taxon=SEPSYMBOL.$taxon;
			}
			$taxon=$name."(".$rank.")".$taxon;
		} else {
			#print "false\n";
		}
		#next ncbi_code:
		$ncbi_code= $parent;
	
		
	}

	return $taxon;

}


#####################################
# Checks if taxa string is well formated
# Remove family, ... 
#####################################
sub check_taxa{
	my ($taxa,$delim) = @_;	
	# help for debugging - search substrings within pfam:
	# "select taxonomy from pfamseq WHERE MATCH (taxonomy) AGAINST ('+blabla' IN BOOLEAN MODE);"
	
	my $original = $taxa;

	$taxa =~ s/^\s+//;
	$taxa =~ s/\s+$//;
	$taxa =~ s/\.$//g;

	$delim = ' ' unless $delim;
		
	die "ERROR: wrong taxa: $taxa\n" unless $taxa =~ /^([A-Z]*[a-z]+)/;

	my $kingdom = $1;
	my $is_prokaryote = 0;
	$is_prokaryote = 1 if $kingdom eq 'Bacteria' or $kingdom eq 'Archaea';

	my $is_eukaryote = 0;
	$is_eukaryote = 1 if $kingdom eq 'Eukaryota';
	
	#####################################
	# Check Bacteria and Archaea
	#####################################
	if($is_prokaryote){
		# ??my $check = 0;


		# example: "Bacteria; Bacteroidetes; Bacteroidetes (class); Bacteroidales;Bacteroidaceae; Bacteroides."
#		if( ($taxa =~ / [a-z]+ /) or ($taxa =~ /complex/) or 
#			($taxa =~ /group/) or 
#			($taxa =~ /class/) or ($taxa =~ /\(/)  )
#		{  
#			$check = 0;
#		}
		
		
		
#		if($original =~ /class/ or ($original =~ /\#/) or 
#			($original =~ /VII/) or 
#			$original =~ /CFB-/ or $original =~ /unidentified/ or 
#			$original =~ /bacterium19/)
#		{
#		$check = 0;
#		}
		
		while($taxa =~ s/([a-z])([A-Z])/$1$delim$2/){		
		}
		
		$taxa =~ s/Anaplasma${delim}phagocytophilum${delim}group/Anaplasma/;

		$taxa =~ s/Ehrlichia${delim}canis${delim}group/Ehrlichia/;

		$taxa =~ s/Rickettsia${delim}spotted${delim}fever${delim}group/Rickettsia/;
		$taxa =~ s/Rickettsia${delim}typhus${delim}group/Rickettsia/;
		$taxa =~ s/Burkholderia${delim}pseudomallei${delim}group/Burkholderia/;
		
		$taxa =~ s/[A-Z][a-z]+?$delim[a-z]+${delim}complex//g;
		$taxa =~ s/[A-Z][a-z]+${delim}complex//g;
		$taxa =~ s/[A-Z][a-z]+\/[A-Z][a-z]+${delim}group//g;
		$taxa =~ s/[A-Z][a-z]+$delim[a-z]+${delim}group//g;
		$taxa =~ s/[A-Z][a-z]+$delim[a-z]+${delim}[a-z]+?${delim}group//g;
		$taxa =~ s/[A-Z][a-z]+${delim}group//g;
		$taxa =~ s/[A-Z][a-z]+${delim}subgroup//g;
		$taxa =~ s/[A-Z][a-z]+${delim}[a-z]+?${delim}subgroup//g;

		$taxa =~ s/$delim\(class\)/:class/g;


		$taxa =~ s/[A-Z][a-z]+?aceae//g;      #  taxonomic family of plants, algae and fungi


		$taxa =~ s/[A-Z][a-z]+?idae$//g;      # families
		$taxa =~ s/[A-Z][a-z]+?idae$delim//g;

		$taxa =~ s/[A-Z][a-z]+?ineae$//g;     #  taxonomic suborder of plants, algae and fungi
		$taxa =~ s/[A-Z][a-z]+?ineae$delim//g;

		# skip tribe
		# tribe—or infrafamily—is a taxonomic rank between family and genus. It is sometimes subdivided into subtribes
		$taxa =~ s/[A-Z][a-z]+?ieae$//g;      # ??
		$taxa =~ s/[A-Z][a-z]+?ieae$delim//g;


		$taxa =~ s/$delim\(MAC\)//g;          # ??

		# example: "Bacteria; Proteobacteria; Alphaproteobacteria; Rickettsiales;SAR11 cluster; Candidatus Pelagibacter."
		$taxa =~ s/SAR11${delim}cluster//g;   

		while($taxa =~ /$delim$delim/){
			$taxa =~ s/$delim$delim/$delim/g;
		}

			# some check ...
		die if $taxa =~ /$delim$delim/;



		my @split = split($delim,$taxa);
		my @keep = ();
		foreach my $t (@split){
			my $firstchar = substr($t,0,1);
			my $secondchar = substr($t,1,1);
		
			if(lc($firstchar) eq $firstchar){
				next;
			}
			if(uc($secondchar) eq $secondchar){
				next;
			}
			push @keep, $t;
		}

		$taxa = join($delim,@keep);
		#print "TAXA: $taxa\n";

		if( ($taxa =~ /$delim[a-z]/) or ($taxa =~ /\(/) ){	
			print STDERR "check: $taxa\n";
		}

		# some checks ...
		die "ERROR: wrong taxa: $taxa\n" if $taxa =~ / complex /;
		die if $taxa =~ / group /;
		die "ERROR: taxa: $taxa : ".uc($taxa)." ".lc($taxa)."\n" 
						 if $taxa =~ /\(class\) /;
		die if $taxa =~ /\(MAC\)/;


	}
	##############################
	# Check taxa of Eukaryotes
	##############################
	elsif($kingdom eq 'Eukaryota'){
	
		$taxa =~ s:$[A-Z][a-z]+/[A-Z][a-z]+${delim}group::g;

		$taxa =~ s/$delim[a-z]+$delim/$delim/g;

		while($taxa =~ s/([a-z])([A-Z])/$1$delim$2/){		
		}

		$taxa =~ s/$delim\w+?inae$delim/$delim/g;
		$taxa =~ s/$delim\w+?inae$//;

		$taxa =~ s/$delim\w+?idae$delim/$delim/g;
		$taxa =~ s/$delim\w+?idae$//g;

		# remove families
		$taxa =~ s/$delim\w+?aceae$delim/$delim/g;
		$taxa =~ s/$delim\w+?aceae$//;

		# remove tribe
		$taxa =~ s/$delim\w+?ini$delim/$delim/g;
		$taxa =~ s/$delim\w+?ini$//;

		$taxa =~ s/${delim}unclassified$delim/$delim/ig;
		$taxa =~ s/unclassified//ig;
		$taxa =~ s/$delim[a-z]+$delim/$delim/g;

		my @split = split($delim,$taxa);
		my @keep = ();
		foreach my $t (@split){
			my $firstchar = substr($t,0,1);
			my $secondchar = substr($t,1,1);

			if(lc($firstchar) eq $firstchar){
				next;
			}
			if(uc($secondchar) eq $secondchar){
				next;
			}
			if($firstchar eq '(' ){
				next;
			}
		  
			push @keep, $t;
		}

		$taxa = join($delim,@keep);

	}
	return $taxa;
}

####################################
# Parse Pfam Results
####################################
sub parse_pfam{
	my @args = @_;

	my $file = shift @args;
	my $subref = shift @args;
	
	unless(open(PFAM,"<$file")){
		print STDERR "ERROR: Could not open file $file\n$!\n";
		exit 1;
	}
	my $line;
	my $state = 0;
	my $query;
	my @hits;
	my $counter;

	my $last_query;
	while($line = <PFAM>){
		chomp $line;
		next if $line =~ /^\s*$/;
	
		# get query name
		if($state == 0){
			next unless $line =~ /^Query sequence:\s+(.*?)\s*$/;
			$state = 1;
			$query = $1;
			next;
		}
		elsif($state == 1){
			next unless $line =~ /^--------/;
			$state = 2;
			next;
		}
		# get family classification
		elsif($state == 2){
			if($line =~ /^\s+\[no hits above thresholds\]/){
				$state = 7;
				@hits = ();
				next;
			}
			# we do not care about domain descriptions
			elsif($line =~ /^Parsed for domains:/){
				$state = 7;

				# make sure to count each seq only once
				unless($last_query and  $last_query eq $query){
					$counter++;

					unless(&$subref($query,\@hits,@args)){
					print STDERR "ERROR while parsing file!\n";
					print STDERR "Terminating parse_pfam ....\n";
					return;
					}
				}
				$last_query = $query;
				@hits = ();
				next;
			}
			else{
				die "ERROR: wrong line \n$line\n" unless $line =~ /^(\S+)\s+(.*?)\s+([\d\.\-]+)\s+([\d\.e\-]+)\s+\d+\s*$/;
				my $acc = $1;
				my $descr = $2;
				my $score = $3;
				my $evalue = $4;
		
				push @hits, [$acc,$descr,$score,$evalue];
			}
		
		}
		# go to end of hit description for query
		elsif($state == 7){
			next unless $line =~ /^\/\//;
			$state = 0;
			next;
		}
	}

	close(PFAM);
}



###########################################
# Get pfam accession number
# by pfam id
###########################################
sub get_pfamACC_by_pfamID{
	my ($pfam_id) = @_;

	unless(ref $STH_ACC){
	unless(ref $DBH){
		init_dbh();
	}
	
	$STH_ACC = $DBH->prepare(q{
		SELECT  pfamA_acc from pfamA where pfamA.pfamA_id = ? 
		}) or die "ERROR: ".$DBI::errstr;
	}

	$STH_ACC->execute($pfam_id);

	my @row;

	my $acc;

	# iterate over each acc term
	while(@row = $STH_ACC->fetchrow_array){
		die if @row > 1;
		# there should be only one accession number
		die if $acc;
		($acc) = @row;
	}
	

	return $acc;
}

###########################################
# Get pfam accession number
# by pfam id
###########################################
sub get_pfamID_by_pfamACC{
	my ($pfam_acc) = @_;

	unless(ref $STH_ID){
	unless(ref $DBH){
		init_dbh();
	}
	
	$STH_ID = $DBH->prepare(q{
		SELECT  pfamA_id from pfamA where pfamA.pfamA_acc = ? 
		}) or die "ERROR: ".$DBI::errstr;
	}

	$STH_ID->execute($pfam_acc);

	my @row;

	my $acc;

	# iterate over each acc term
	while(@row = $STH_ID->fetchrow_array){
		die if @row > 1;
		# there should be only one accession number
		die if $acc;
		($acc) = @row;
	}
	

	return $acc;
}


#########################################
# Get GO term for Pfam family
#########################################
sub get_go_by_pfam_name{
	my ($name) = @_;

	unless(ref $STH_GO){
	unless(ref $DBH){
		init_dbh();
	}
	
	$STH_GO = $DBH->prepare(q{
		SELECT go_id,term,category from pfamA,gene_ontology where pfamA.pfamA_id = ? and pfamA.auto_pfamA = gene_ontology.auto_pfamA
		}) or die "ERROR: ".$DBI::errstr;
	}

	$STH_GO->execute($name);
	
	my @results;
	my @row;

	# iterate over each go term
	while(@row = $STH_GO->fetchrow_array){
		my $term = $row[1];
		my $class = $row[2];
		$term =~ s/\,//g;
		push @results, {term => $term, class => $class};
	}
	

	if($name eq 'CytoC_RC' or $name eq 'BLUF' or $name eq 'Lycopene_cycl' or $name eq 'Cytochrom_B559'  or $name eq 'Cytochrom_B559a' or $name eq 'LHC' or $name eq 'Apocytochr_F_C' or $name eq 'GFP' or $name eq 'CpcD' or $name eq 'YCF9' or $name eq 'PSI_PsaE' or $name eq 'PetG' or $name eq 'PetN' or $name eq 'PRCH' or $name eq 'PetL' or $name eq 'PsbM' or $name eq 'PRC' or $name eq 'PsbU' or $name eq 'PetM'){
		push @results, {term => 'photosynthesis', class => 'process'};
		#print " $name\n";
	}

	return \@results;
}


sub get_go_id_by_pfam_acc{
	my ($name) = @_;

	unless(ref $STH_GO){
	unless(ref $DBH){
		init_dbh();
	}
	
	$STH_GO = $DBH->prepare(q{
		SELECT go_id from pfamA,gene_ontology where pfamA.pfamA_acc = ? and pfamA.auto_pfamA = gene_ontology.auto_pfamA
		}) or die "ERROR: ".$DBI::errstr;
	}

	$STH_GO->execute($name);
	
	my @results;
	my @row;

	# iterate over each go term
	while(@row = $STH_GO->fetchrow_array){
		my $go_id=$row[0];
		#my $term = $row[1];
		#my $ontology = $row[2];

		#$term =~ s/\,//g;
		#push @results, {term => $term, class => $class};
		push(@results, $go_id);
	}
	

#	if($name eq 'CytoC_RC' or $name eq 'BLUF' or $name eq 'Lycopene_cycl' or $name eq 'Cytochrom_B559'  or $name eq 'Cytochrom_B559a' or $name eq 'LHC' or $name eq 'Apocytochr_F_C' or $name eq 'GFP' or $name eq 'CpcD' or $name eq 'YCF9' or $name eq 'PSI_PsaE' or $name eq 'PetG' or $name eq 'PetN' or $name eq 'PRCH' or $name eq 'PetL' or $name eq 'PsbM' or $name eq 'PRC' or $name eq 'PsbU' or $name eq 'PetM'){
#		push @results, {term => 'photosynthesis', class => 'process'};
		#print " $name\n";
#	}

	return \@results;
}


