package Tree;

#
#  Copyright (C) 2007 CeBiTec, Bielefeld University
#  Written by Lutz Krause (Lutz.Krause@CeBiTec.Uni-Bielefeld.DE)
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  version 2 as published by the Free Software Foundation.
#
#  This file is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this file; see the file LICENSE.  If not, write to
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.
#

use strict;
use warnings;

use TCommon;
use TConfigs;

use constant MIN_LEAVE_NUMBER => 3;

no warnings "recursion";

1;

##############################
# Create new Tree instance
##############################
sub new{
    my ($class) = @_;

    my $self = {};

    bless $self, $class;

    return $self;
}

############################### 
# Parses file with id,acc tupels
# and returns ref on hash
# with acc as key and respective
# id as value
############################### 
sub get_id_by_acc{
    my ($file) = @_;

    unless(open(IN,"<$file")){
	print STDERR "ERROR: Could not open file $file\n";
	die $!;
    }
    my $line;
    my %acc_by_id = ();
    while($line = <IN>){
	next if $line =~ /^\s*$/;
	
	die unless $line =~ /(\S+)\s+(\S+)/;
	$acc_by_id{$1} = $2;
    }

    close(IN);

    return \%acc_by_id;
}

#############################################
# Parses tree file in newick format
# Returns: Root node and list of leaves
# representing environmental DNA fragments
#############################################
sub get_tree{
    my ($file,$id_acc_file) = @_;

    my $tree;

    my $acc_by_id = get_id_by_acc($id_acc_file);

    unless(open(IN,"<$file")){
	print STDERR "ERROR: Could not open file $file:\n$!\n";
    }

    my @open_nodes = ();

    my @all_nodes;

    my @fragment_nodes = ();
    
    my $line;
    

    # mode:
    # 0: begin node
    # 1 get acc
    # 2 get distance of leaf
    # 4 get distance of node
    my $mode = 0;

    my $acc;
    my $dist;

    my $id = 0;

    my $pushed = 0;
    my $poped = 0;

    #########################
    # Parse newick tree
    #########################
    while($line = <IN>){
	chomp($line);
	next if $line =~ /^\s*$/;
	
	while($line =~ /(.)/g){
		my $char = $1;
		
		# open node
		if($char eq '('){		   
		    # create new inner node
		    $id++;
		    my $inner_node = {a => $id};
		    		    
		    if(@open_nodes){
			my $last_index = @open_nodes - 1;
			my $parent = $open_nodes[$last_index];
			$inner_node->{p} = $parent;
			push @{ $parent->{c}}, $inner_node;
		    }
		    
		    push @open_nodes, $inner_node;
		    $pushed++;
		    
		    $mode = 1;
		    
		}
		elsif($char eq ','){		    
		    # add new node
		    _add_node(\@open_nodes,\@fragment_nodes,\$acc,\$dist,$mode,\$poped,$acc_by_id);
		    
		    $mode = 1;
		}
		elsif($char eq ')'){
		    # add new node
		    _add_node(\@open_nodes,\@fragment_nodes,\$acc,\$dist,$mode,\$poped,$acc_by_id);
		    
		    $mode = 4;
		}
		elsif($char eq ';'){
		    die unless $mode == 4;
		    
		    my $n = @open_nodes;
		    die "ERROR: open nodes $n ($pushed,$poped)\n" unless @open_nodes == 1;

		    my $root = pop @open_nodes;

		    close(IN);

		    # return root node and list of leaves
		    # representing environmental
		    # DNA fragments
		    return ($root,\@fragment_nodes);
		}
		elsif($mode == 4){
		    next if $char eq ':';

		    $dist .= $char;
		} 		    
		elsif($mode == 2){
		    $dist .= $char;
		}
		elsif($mode == 1){
		    if($char eq ':'){
			$mode = 2;
			next;
		    }		    
		    $acc .= $char;
		}
	}	    	    
    }
    # we should never reach here
    die;
}

#################################
# add new node to tree
#################################
sub _add_node{
    my ($open_nodes,$fragment_nodes,$acc,$dist,$mode,$poped,$acc_by_id) = @_;

    # add inner node
    #
    if($mode == 4){
	my $open_node = pop @$open_nodes;
	$$poped++;
	$open_node->{d} = $$dist;
	$$acc = undef;
	$$dist = undef;
    }

    #add leaf
    #
    elsif($mode == 2){
	my $last_index = @$open_nodes - 1;
	my $parent = $open_nodes->[$last_index];
	
	my $new_leaf = {d => $$dist, p => $parent};		    
	
	push @{ $parent->{c} }, $new_leaf;
	my $a = $acc_by_id->{$$acc};
	

	die unless $a;

	$new_leaf->{a} = $a;
	
	unless(TCommon::_is_pfam($a)){
	    push @$fragment_nodes,$new_leaf;
	}
	$$acc = undef;
	$$dist = undef;
    }   
    
    return 1;
}


#############################################
# Predict taxonomic origin of environmental
# DNA fragment
#############################################
sub get_fragment_taxa{
    my ($fragment,$file) = @_;

    my $parent1 = $fragment->{p};
    
    my $leaves1;
    
    # array containing tuple of parents/leaves of fragment
    my @parents = ();

    # get parent node that has leaves with known taxonomic affiliation
    while(1){	
	my $i = 0;
	$leaves1 = get_all_child_pfam_leaves($parent1,\$i,"1$file");

	push @parents,[$parent1,$leaves1];
	
	
	# found parent with leaves
	if(@$leaves1){
	    last;
	}
	# get new parent 1 
	else{
	    if(exists $parent1->{p}){
		$parent1 = $parent1->{p};
		next;
	    }
	    else{
		$parent1 = undef;
		last;
	    }
	    
	}
    }
    
    unless(ref $parent1){
	# this should never happen, at least root should have leaves!!!
	die "ERROR: fragment $fragment->{a} has no parent with leaf!!!\n";
    }
    unless(exists $parent1->{p}){
	return;
    }
    
    my $parent2 = $parent1->{p};
    my $leaves2;

    # get parent node that
    # has at least MIN_LEAVE_NUMBER child leaves with
    # known taxonomic affiliation
    while(1){	
	my $i = 0;
	
	# get all child leaves
	$leaves2 = get_all_child_pfam_leaves($parent2,\$i,"2$file");
	
	push @parents,[$parent2,$leaves2];

	# found parent with more leaves that parent1
	if((@$leaves2 > @$leaves1) and @$leaves2 >= MIN_LEAVE_NUMBER){
	    last;
	}
	# get new parent 1 
	else{
	    if(exists $parent2->{p}){
		$parent2 = $parent2->{p};
		next;
	    }
	    else{
		$parent2 = undef;
		last;
	    }
	    
	}
    }
    
    unless(ref $parent2){
	return;
    }

    # for each taxonomic level (superkingdom, phylum, ...)
    # get dominant taxa for 
    # subtree of parent
    my %dominant_taxa = ();

    foreach my $tuple (@parents){
	my ($parent,$leaves) = @$tuple;
	
	my $dominant_taxa_parent = get_dominant_taxa($parent,$leaves,$file);
	
	next unless keys %$dominant_taxa_parent;

	foreach my $level (keys %$dominant_taxa_parent){
	    $dominant_taxa{$level}->{$dominant_taxa_parent->{$level}} = 1;
	}
    }
        
    my $taxa = '';
    my $dt = '';
    foreach my $level (sort {$a <=> $b} keys %dominant_taxa){
	my @tmp = keys %{ $dominant_taxa{$level} };
       
	if(@tmp == 1){
	    
	    $dt = shift @tmp;

	    # check if leaf is embedded in taxa
	    if(is_embedded_in($fragment,$dt,$file)){
		$taxa = $dt;
	    }
	    else{
		return $taxa;
	    }
	}
	else{
	    return $taxa;
	}
    }
    return $taxa;
}

##############################
# Get taxonomic origin of 
# Pfam family member
##############################
sub get_pfam_taxa{
    my ($leaf) = @_;
    
    my $acc = $leaf->{a};
  return TCommon::get_taxa($acc);

#    return $leaf->{t} if exists $leaf->{t};

 #   my $taxa = TCommon::get_taxa($acc);
        
 #   $leaf->{t} = $taxa;

 #   return $taxa;
}


##############################
# Get all child leaves of inner
# node
##############################
sub get_all_child_pfam_leaves{
    my ($node,$i,$file) = @_;
   
    $$i++;


    my @all_leaves = ();
    if(exists $node->{c}){
	my $childs = $node->{c};
	
	foreach my $child (@$childs){
	    if(exists $child->{c}){
		my $child_leaves = get_all_child_pfam_leaves($child,$i,$file);
		push @all_leaves,@$child_leaves;
	    }
	    elsif(TCommon::_is_pfam($child->{a})){
		
		push @all_leaves,$child;
	    }
	}
    }

    $$i--;
    
    return \@all_leaves;
}

##############################
# get dominant taxa in subtree
# of $node
# leaves: leaves of subtree
# this parameter is optional
#
# returns undef if subtree does
# have leaves representing Pfam
# family members
##############################
sub get_dominant_taxa{
    my ($node,$leaves,$file) = @_;

#    my $sep = SEPSYMBOL;
    my $leaves2;

    if(defined $leaves){
	$leaves2 = $leaves;
    }
    else{
	my $i = 0;
	$leaves2 = get_all_child_pfam_leaves($node,\$i,"4$file");
    } 

    my %common = ();

    return unless @$leaves;

    foreach my $leaf (@$leaves2){
	my $tax = get_pfam_taxa($leaf);
#print join(",", @$tax)."\n";

	unless((ref $tax) and @$tax){
	    print STDERR "WARNING (get_dominant_taxa): no tax for leaf ",$leaf->{a},"\n";
	    next;
	}

	next if $tax =~ /other/;
	next if $tax =~ /unidentified/;

	my $level = 0;
	my $joined;
	foreach (@$tax){
	    if($joined){
		
	    	if ($_ ne "") {
			$joined .= SEPSYMBOL . $_;
		}

	    }
	    else{
		$joined = $_;

	    }
	    $common{$level}->{$joined}++;
	    $level++;
	}
#print "J: ".$joined."\n"	;
	
    }
    my ($tax,$count);
    
        
    my $node_acc = $node->{a};
    
    my %dominant_taxa = ();

    foreach my $level (sort {$a <=> $b} keys %common){

	my $max_count = 0;
	my $max_taxa = 'bla';
	my $second_max = 0;
	
	my $total_count = 0;
	while( ($tax,$count) = each %{ $common{$level} }){
	    $total_count++;
	    
	    if($count >= $max_count){
		$second_max = $max_count;
		$max_count = $count;	    
		$max_taxa = $tax;
	    }
	    elsif($count > $second_max){
		$second_max = $count;
	    }
	}

	# check if at least 80% of members of
	# subtree share a common taxon
	if($max_count > 0.8  * $total_count){  
	    $dominant_taxa{$level} = $max_taxa;
	    #print "max: $level, $max_taxa\n";
	}
	else{
	    return \%dominant_taxa;
	}
    }
    return \%dominant_taxa;
}


##############################
# Checks if the subtree arising
# from $level2_parent_node
# has a child node with 
# taxon $tax
##############################
sub is_embedded_in{
    my ($level2_parent_node,$tax,$file) = @_;

    my $children = $level2_parent_node->{c};

    foreach my $child (@$children){
	my $has_tax = 0;
	
	# child is inner node
	if(exists $child->{c}){	    
	    my $i = 0;
	    my $sub_leaves = get_all_child_pfam_leaves($child,$i,"5$file"); 
	    
	    foreach my $leaf (@$sub_leaves){
		my $leaf_taxa = get_pfam_taxa($leaf);
		

		$leaf_taxa = join(SEPSYMBOL,@$leaf_taxa);

		
		if($leaf_taxa =~ /^$tax/){
		    $has_tax = 1;
		    last;
		}	    
	    }
	    return unless $has_tax;
	}
	# child is a leaf
	else{
	    my $tmp_tax = join(SEPSYMBOL,@{ get_pfam_taxa($child) });

	    return unless $tmp_tax =~ /^$tax/;
	}
    }
    return 1;
}
